(*  Title:      HOL/Eisbach/Eisbach_Tools2.thy
    Author:     Daniel Matichuk, NICTA/UNSW
                Frederic Tuong, Virginia Tech
                Yakoub Nemouchi, Virginia Tech

Extended version of Eisbach tools.
*)

theory Eisbach_Tools2
  imports "~~/src/HOL/Eisbach/Eisbach_Tools"
          "~~/src/HOL/Code_Evaluation"
begin

ML\<open>
(*  Title:      HOL/Tools/value_command.ML
    Author:     Florian Haftmann, TU Muenchen

Generic value command for arbitrary evaluators, with default using nbe or SML.
*)

structure Value_Command' =
struct

fun default_value ctxt t =
  if null (Term.add_frees t [])
  then Code_Evaluation.dynamic_value_strict ctxt t
  else Nbe.dynamic_value ctxt t;
end
\<close>

ML \<open>
local

fun trace_method parser print =
  Scan.lift (Args.mode "dummy") -- parser >>
    (fn (dummy, x) => fn ctxt => SIMPLE_METHOD (fn st =>
      (if dummy orelse not (Method.detect_closure_state st)
       then tracing (print ctxt x) else ();
       all_tac st)));

fun setup_trace_method0 binding show_msg parser print =
  Method.setup binding
    (trace_method parser
      (fn ctxt => fn x =>
        (if show_msg then Markup.markup Markup.keyword3 (Binding.name_of binding ^ ": ") else "")
        ^ print ctxt x))
    "";

fun setup_trace_method binding = setup_trace_method0 binding true

fun unparse tok = Markup.markup Markup.keyword2 (Token.unparse tok ^ ": ")

fun intensify s = s |> Markup.markup Markup.inner_cartouche |> Markup.markup Markup.intensify;

fun intensify' s = s |> Markup.markup (Markup.bad ());

val Args_text = Scan.peek (K Args.text)

fun setup_trace_method' binding = setup_trace_method binding Args_text (K intensify)

in

val _ =
  Theory.setup
   (setup_trace_method @{binding print_fact}
      (Scan.lift (Scan.ahead Parse.not_eof) -- Attrib.thms)
      (fn ctxt => fn (tok, thms) =>
        unparse tok ^ commas (map (Thm.string_of_thm ctxt) thms)) #>
    setup_trace_method @{binding print_fact'}
      Attrib.thms
      (fn ctxt => fn thms => commas (map (Thm.string_of_thm ctxt) thms)) #>
    setup_trace_method @{binding print_term}
      (Scan.lift (Scan.ahead Parse.not_eof) -- Args.term)
      (fn ctxt => fn (tok, t) => unparse tok ^ Syntax.string_of_term ctxt t) #>
    setup_trace_method @{binding print_term'}
      Args.term
      (fn ctxt => fn t =>
        Syntax.string_of_term ctxt (Value_Command'.default_value ctxt t)) #>
    setup_trace_method0 @{binding print_string} false Args_text (K intensify) #>
    setup_trace_method0 @{binding print_error} false Args_text (K intensify') #>
    setup_trace_method' @{binding computation_mode} #>
    setup_trace_method @{binding formalism_used_is}
      Args_text
      (K (K (intensify "Unifying Theories of Programming (UTP)"))) #>
    setup_trace_method' @{binding reasoning} #>
    setup_trace_method' @{binding correctness} #>
    setup_trace_method @{binding print_type'} Args.typ Syntax.string_of_typ);

end
\<close>

end