(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)

theory algorithms

  imports "../Backend/VCG/vcg"
     "../Midend-IVL/Isabelle-UTP-Extended/utils/vcg_helpers" 
begin
recall_syntax \<comment> \<open>Fixes notation issue with inclusion of HOL libraries.\<close>

section \<open>setup and makeup!\<close>    
declare utp_lift.lift_pre_var [simp del]

sledgehammer_params[stop_on_first,parallel_subgoals, join_subgoals]
  
no_adhoc_overloading'_all

section \<open>Total vs Partial\<close>

lemma 
 "\<lbrace>true\<rbrace>
   INVAR true
   VRT  \<guillemotleft>0::nat\<guillemotright>
   WHILE true DO SKIP OD
  \<lbrace>false\<rbrace>\<^sub>P"
  apply (vcg wp)  
  oops    

lemma 
 "\<lbrace>true\<rbrace> 
  invr true while\<^sup>\<top> true do SKIP\<^sub>r od
  \<lbrace>false\<rbrace>\<^sub>u"
  apply (vcg wp)                           
  done

section \<open>Simple algorithms\<close>

text 
\<open>
 Through these experiments I want to observe the following problems:
     \begin{itemize}
       \item I want to deal with the problem of nested existential(SOLVED).
       \item I want to deal with the problem of blow up due to the semantic machinery coming with lenses(SOLVED).
       \item I want to have modularity(NOT SOLVED).
     \end{itemize}
\<close>
subsection \<open>addition method\<close>

lemma add_prog_partial: 
  assumes "vwb_lens a" "vwb_lens b"
  assumes "a \<bowtie> b"
  shows  
 "\<lbrace>\<guillemotleft>y\<guillemotright> >\<^sub>u \<guillemotleft>0\<guillemotright>\<rbrace>
   assign_r a  \<guillemotleft>x\<guillemotright> ;;
   assign_r b  \<guillemotleft>y::nat\<guillemotright>;;
   invr &a =\<^sub>u  \<guillemotleft>x\<guillemotright> + (\<guillemotleft>y\<guillemotright> - &b)  \<and>  \<guillemotleft>y\<guillemotright> \<ge>\<^sub>u &b \<and>
         &b =\<^sub>u  (\<guillemotleft>y\<guillemotright> + &b)  - \<guillemotleft>y\<guillemotright> 
   while\<^sup>\<top> &b \<noteq>\<^sub>u \<guillemotleft>0\<guillemotright>
     do assign_r a (&a + \<guillemotleft>1\<guillemotright>) ;; 
        assign_r b (&b - \<guillemotleft>1\<guillemotright>)
     od
  \<lbrace>&a =\<^sub>u \<guillemotleft>y\<guillemotright> + \<guillemotleft>x\<guillemotright> \<rbrace>\<^sub>u"
  apply (insert assms) (*Make this automatic *)
  apply (vcg wp) 
  done

term "e\<lbrakk>&b - \<guillemotleft>1\<guillemotright>/(&b)\<rbrakk> "
term"subst ([b \<mapsto>\<^sub>s (&b - \<guillemotleft>1\<guillemotright>)]) ((&\<Sigma>, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>measure \<lbrakk>&b\<rbrakk>\<^sub>e\<guillemotright>) = XX
      "
lemma "vwb_lens b \<Longrightarrow> subst ([b \<mapsto>\<^sub>s (&b - \<guillemotleft>1\<guillemotright>)]) (&b) = (&b - \<guillemotleft>1\<guillemotright>)"
  apply  (simp add: usubst)
  
  oops
  term "\<guillemotleft>measure (\<lambda>_. \<lbrakk>&b\<rbrakk>\<^sub>e st)\<guillemotright>"
lemma "subst ([b \<mapsto>\<^sub>s (&b - \<guillemotleft>1\<guillemotright>)]) ((&\<Sigma>, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>measure \<lbrakk>&b\<rbrakk>\<^sub>e\<guillemotright>) = XX"
  apply  (simp add: usubst)
 
  oops
  term "\<langle> ([X \<mapsto>\<^sub>s v])\<rangle>\<^sub>s X"
lemma "a \<bowtie> b \<Longrightarrow> vwb_lens b \<Longrightarrow> vwb_lens a \<Longrightarrow>`(&b) \<noteq>\<^sub>u \<guillemotleft>0\<guillemotright> \<and>
            (&a =\<^sub>u \<guillemotleft>x\<guillemotright> + (\<guillemotleft>y\<guillemotright> - (&b)) \<and> bop op \<le> (&b) \<guillemotleft>y\<guillemotright>) \<and> &\<Sigma> =\<^sub>u \<guillemotleft>st\<guillemotright> \<Rightarrow>
            ((&\<Sigma>, \<guillemotleft>st\<guillemotright>)\<^sub>u \<in>\<^sub>u \<guillemotleft>measure \<lbrakk>&b \<rbrakk>\<^sub>e\<guillemotright>)\<lbrakk>&b - \<guillemotleft>1\<guillemotright>/(&b)\<rbrakk>\<lbrakk>&a + \<guillemotleft>1\<guillemotright>/(&a)\<rbrakk>`"
  apply (subst utp_subst.subst_bop)+
  using [[simp_trace]]
  apply  (simp add: usubst)
    
  oops
lemma add_prog_total:
  assumes "vwb_lens a" "vwb_lens b"
  assumes "a \<bowtie> b"
  shows  
 "\<lbrace>(\<guillemotleft>y::nat\<guillemotright>) >\<^sub>u 0\<rbrace>
   a :== \<guillemotleft>x\<guillemotright> ;
   b :== \<guillemotleft>y\<guillemotright>;
   INVAR  &a =\<^sub>u  \<guillemotleft>x\<guillemotright> + (\<guillemotleft>y\<guillemotright> - &b)  \<and>  \<guillemotleft>y\<guillemotright> \<ge>\<^sub>u &b 
   VRT (&b)
   WHILE &b \<noteq>\<^sub>u \<guillemotleft>0\<guillemotright>  
     DO a:== (&a + \<guillemotleft>1\<guillemotright>);
        b :== (&b - \<guillemotleft>1\<guillemotright>)
     OD
  \<lbrace>&a =\<^sub>u \<guillemotleft>y\<guillemotright> + \<guillemotleft>x\<guillemotright>\<rbrace>\<^sub>P"
  apply (insert assms) (*Make this automatic *)
  apply (vcg wp)
  done

subsection \<open>Increment method\<close>
  
lemma increment_method_sp_H1_H3: 
  assumes "vwb_lens x"
  shows  
 "\<lbrace>\<guillemotleft>a\<guillemotright> >\<^sub>u 0\<rbrace>
   x :== 0 ; 
   INVAR \<guillemotleft>a\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u &x 
   VRT   uop nat (\<guillemotleft>a\<guillemotright> - &x) (*Change to the newest pattern*)
   WHILE &x <\<^sub>u \<guillemotleft>a\<guillemotright> DO x:== (&x + 1) OD
  \<lbrace>\<guillemotleft>a\<guillemotright> =\<^sub>u &x\<rbrace>\<^sub>P"
  apply (insert assms) (*Make this automatic *)
  apply (vcg sp)             
  done

lemma increment_method_wp_H1_H3: 
  assumes "vwb_lens x"
  shows  
 "\<lbrace>\<guillemotleft>a\<guillemotright> >\<^sub>u 0\<rbrace>
   x :== 0 ; 
   INVAR \<guillemotleft>a\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u &x 
   VRT uop nat (\<guillemotleft>a\<guillemotright> - &x)
   WHILE &x <\<^sub>u \<guillemotleft>a\<guillemotright> DO x:== (&x + 1) OD
  \<lbrace>\<guillemotleft>a\<guillemotright> =\<^sub>u &x\<rbrace>\<^sub>P"
  apply (insert assms) (*Make this automatic *)
  apply (vcg wp)                           
  done

lemma increment_method_sp_rel: 
  assumes "vwb_lens x"
  shows  
 "\<lbrace>\<guillemotleft>a:: nat\<guillemotright> >\<^sub>u 0\<rbrace>
   assign_r x  0 ;;
   invr \<guillemotleft>a\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u &x 
   while\<^sup>\<top> &x <\<^sub>u \<guillemotleft>a\<guillemotright> do assign_r x (&x + 1) od
  \<lbrace>\<guillemotleft>a\<guillemotright> =\<^sub>u &x\<rbrace>\<^sub>u"
  apply (insert assms) (*Make this automatic *)
  apply (vcg sp)                           
  done
    
lemma increment_method_wp_rel: 
  assumes "vwb_lens x"
  shows  
 "\<lbrace>\<guillemotleft>a:: nat\<guillemotright> >\<^sub>u 0\<rbrace>
   assign_r x  0 ;;
   invr \<guillemotleft>a\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u &x 
   while\<^sup>\<top> &x <\<^sub>u \<guillemotleft>a\<guillemotright> do assign_r x (&x + 1) od
  \<lbrace>\<guillemotleft>a\<guillemotright> =\<^sub>u &x\<rbrace>\<^sub>u"
  apply (insert assms) (*Make this automatic *)
  apply (vcg wp)                           
  done
       
subsection \<open>even count program\<close> 

lemma even_count_gen_sp_H1_H3:
  assumes "i \<bowtie> j"
  assumes "vwb_lens i" "vwb_lens j"  
  shows  
 "\<lbrace>\<guillemotleft>a\<guillemotright> >\<^sub>u 0 \<rbrace>
   i :== \<guillemotleft>0::int\<guillemotright>;
   j :== 0 ; 
   INVAR (&j =\<^sub>u (&i + 1) div \<guillemotleft>2\<guillemotright> \<and> &i \<le>\<^sub>u \<guillemotleft>a\<guillemotright>) 
   VRT   uop nat (\<guillemotleft>a\<guillemotright> - &i)
   WHILE &i <\<^sub>u \<guillemotleft>a\<guillemotright>
   DO
     IF &i mod \<guillemotleft>2\<guillemotright> =\<^sub>u 0 
     THEN j :== (&j + 1)
     ELSE SKIP 
     FI;
     i :== (&i + 1)
   OD
  \<lbrace>&j =\<^sub>u (\<guillemotleft>a\<guillemotright> + 1)div \<guillemotleft>2\<guillemotright>\<rbrace>\<^sub>P" 
  apply (insert assms)(*Make this automatic*)
  apply (vcg wp) 
   apply presburger+    
  done   

lemma even_count_gen'_sp_H1_H3:
  assumes "i \<bowtie> j"
  assumes "vwb_lens i" "vwb_lens j"  
  shows  
 "\<lbrace>\<guillemotleft>a\<guillemotright> >\<^sub>u 0\<rbrace>
   i :== \<guillemotleft>0::int\<guillemotright>;
   j :== 0 ; 
   INVAR  (&j =\<^sub>u (&i + 1) div 2 \<and> &i \<le>\<^sub>u \<guillemotleft>a\<guillemotright>) 
   VRT uop nat (\<guillemotleft>a\<guillemotright> - &i)
   WHILE &i <\<^sub>u \<guillemotleft>a\<guillemotright>
   DO
     IF &i mod 2 =\<^sub>u 0 
     THEN j :== (&j + 1)
     ELSE SKIP 
     FI;
    i :== (&i + 1)
   OD
 \<lbrace>&j =\<^sub>u (\<guillemotleft>a\<guillemotright> + 1)div 2\<rbrace>\<^sub>P"  
  apply (insert assms)(*Make this automatic*)
  apply (vcg sp) 
    apply (simp_all add: zdiv_zadd1_eq)
  done   
    
lemma even_count_gen'_wp_H1_H3:
  assumes "i \<bowtie> j"
  assumes "vwb_lens i" "vwb_lens j"  
  shows  
 "\<lbrace>\<guillemotleft>a\<guillemotright> >\<^sub>u 0\<rbrace>
   i :== \<guillemotleft>0::int\<guillemotright>;
   j :== 0 ; 
   INVAR  (&j =\<^sub>u (&i + 1) div 2 \<and> &i \<le>\<^sub>u \<guillemotleft>a\<guillemotright>) 
   VRT uop nat (\<guillemotleft>a\<guillemotright> - &i)
   WHILE &i <\<^sub>u \<guillemotleft>a\<guillemotright>
   DO
     IF &i mod 2 =\<^sub>u 0 
     THEN j :== (&j + 1)
     ELSE SKIP 
     FI;
    i :== (&i + 1)
   OD
 \<lbrace>&j =\<^sub>u (\<guillemotleft>a\<guillemotright> + 1)div 2\<rbrace>\<^sub>P"  
   apply (insert assms)(*Make this automatic*)
   apply (vcg wp)    
     apply simp_all
   using dvd_imp_mod_0 odd_succ_div_two 
   apply blast
   done       
     
lemma even_count_gen'_sp_rel:
  assumes "i \<bowtie> j"
  assumes "vwb_lens i" "vwb_lens j"  
  shows  
 "\<lbrace>\<guillemotleft>a\<guillemotright> >\<^sub>u 0\<rbrace>
   assign_r i  \<guillemotleft>0::int\<guillemotright>;;
   assign_r j 0 ;; 
   invr  (&j =\<^sub>u (&i + 1) div 2 \<and> &i \<le>\<^sub>u \<guillemotleft>a\<guillemotright>) 
   while\<^sup>\<top> &i <\<^sub>u \<guillemotleft>a\<guillemotright>
   do
     bif &i mod 2 =\<^sub>u 0 
     then assign_r j  (&j + 1)
     else SKIP\<^sub>r 
     eif;;
    assign_r i  (&i + 1)
   od
 \<lbrace>&j =\<^sub>u (\<guillemotleft>a\<guillemotright> + 1)div 2\<rbrace>\<^sub>u"  
  apply (insert assms)(*Make this automatic*)
  apply (vcg sp)    
   apply (simp_all add: zdiv_zadd1_eq)
  done    
    
lemma even_count_gen'_wp_rel:
  assumes "i \<bowtie> j"
  assumes "vwb_lens i" "vwb_lens j"  
  shows  
 "\<lbrace>\<guillemotleft>a\<guillemotright> >\<^sub>u 0\<rbrace>
   assign_r i  \<guillemotleft>0::int\<guillemotright>;;
   assign_r j 0 ;; 
   invr  (&j =\<^sub>u (&i + 1) div 2 \<and> &i \<le>\<^sub>u \<guillemotleft>a\<guillemotright>) 
   while\<^sup>\<top> &i <\<^sub>u \<guillemotleft>a\<guillemotright>
   do
     bif &i mod 2 =\<^sub>u 0 
     then assign_r j  (&j + 1)
     else SKIP\<^sub>r 
     eif;;
    assign_r i  (&i + 1)
   od
 \<lbrace>&j =\<^sub>u (\<guillemotleft>a\<guillemotright> + 1)div 2\<rbrace>\<^sub>u"  
  apply (insert assms)(*Make this automatic*)
  apply (vcg wp)    
  apply simp_all
   using dvd_imp_mod_0 odd_succ_div_two 
   apply blast
  done 
         
subsection \<open>sqrt program\<close>
  
definition Isqrt :: "int \<Rightarrow> int \<Rightarrow> bool" 
   where "Isqrt n\<^sub>0 r \<equiv> 0\<le>r \<and> (r-1)\<^sup>2 \<le> n\<^sub>0" 
     
lemma Isqrt_aux:
  "0 \<le> n\<^sub>0 \<Longrightarrow> Isqrt n\<^sub>0 1"
  "\<lbrakk>0 \<le> n\<^sub>0; r * r \<le> n\<^sub>0; Isqrt n\<^sub>0 r\<rbrakk> \<Longrightarrow> Isqrt n\<^sub>0 (r + 1)"
  "\<lbrakk>0 \<le> n\<^sub>0; \<not> r * r \<le> n\<^sub>0; Isqrt n\<^sub>0 r\<rbrakk> \<Longrightarrow> (r - 1)\<^sup>2 \<le> n\<^sub>0 \<and> n\<^sub>0 < r\<^sup>2"
  "Isqrt n\<^sub>0 r \<Longrightarrow> r * r \<le> n\<^sub>0 \<Longrightarrow> r\<le>n\<^sub>0"
  "\<lbrakk>0 \<le> n\<^sub>0; \<not> r * r \<le> n\<^sub>0; Isqrt n\<^sub>0 r\<rbrakk> \<Longrightarrow> 0 < r"
  apply (auto simp: Isqrt_def power2_eq_square algebra_simps)
  by (smt combine_common_factor mult_right_mono semiring_normalization_rules(3))
     
lemma sqrt_prog_correct_sp_H1_H3:
  assumes "vwb_lens r"
  shows
 "\<lbrace>0 \<le>\<^sub>u \<guillemotleft>a\<guillemotright>\<rbrace>
   r :== 1 ; 
   INVAR 0\<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> bop Isqrt \<guillemotleft>a\<guillemotright> (&r)
   VRT  uop nat ((\<guillemotleft>a\<guillemotright> + 1) - &r)
   WHILE (&r * &r \<le>\<^sub>u \<guillemotleft>a\<guillemotright>)
   DO 
    r :== (&r + 1)
   OD;
   r :== (&r - 1)
 \<lbrace>0\<le>\<^sub>u &r \<and> uop power2 (&r) \<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> \<guillemotleft>a\<guillemotright> <\<^sub>u uop power2 (&r + 1)\<rbrace>\<^sub>P" 
  apply (insert assms)
  supply Isqrt_aux [simp]
  apply (vcg sp)    
  done    

lemma
  assumes "vwb_lens r"
  shows
 "\<lbrace>0 \<le>\<^sub>u \<guillemotleft>a\<guillemotright>\<rbrace>
   r :== 1 ; 
   INVAR 0\<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> uop power2 (&r - 1) \<le>\<^sub>u \<guillemotleft>a\<guillemotright>
   VRT  uop nat ((\<guillemotleft>a\<guillemotright> + 1) - &r )
   WHILE (&r * &r \<le>\<^sub>u \<guillemotleft>a\<guillemotright>)
   DO 
    r :== (&r + 1)
   OD;
   r :== (&r - 1)
 \<lbrace> uop power2 (&r) \<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<rbrace>\<^sub>P" 
  apply (insert assms)
  supply Isqrt_aux [simp]
  apply (vcg sp)    
  apply (simp add: power2_eq_square)
  apply (smt mult_cancel_left1 mult_left_mono)
  done

lemma sqrt_prog_correct_wp_H1_H3:
  assumes "vwb_lens r"
  shows
 "\<lbrace>0 \<le>\<^sub>u \<guillemotleft>a\<guillemotright>\<rbrace>
   r :== 1 ; 
   INVAR 0\<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> bop Isqrt \<guillemotleft>a\<guillemotright> (&r)
   VRT  uop nat ((\<guillemotleft>a\<guillemotright> + 1) - &r)
   WHILE (&r * &r \<le>\<^sub>u \<guillemotleft>a\<guillemotright>)
   DO 
    r :== (&r + 1)
   OD;
   r :== (&r - 1)
 \<lbrace>0\<le>\<^sub>u &r \<and> uop power2 (&r) \<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> \<guillemotleft>a\<guillemotright> <\<^sub>u uop power2 (&r + 1)\<rbrace>\<^sub>P" 
  apply (insert assms)
  supply Isqrt_aux [simp]
  apply (vcg wp)    
  done    
 
lemma sqrt_prog_correct_sp_rel:
  assumes "vwb_lens r"
  shows
 "\<lbrace>0 \<le>\<^sub>u \<guillemotleft>a\<guillemotright>\<rbrace>
   assign_r r 1 ;; 
   invr 0\<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> bop Isqrt \<guillemotleft>a\<guillemotright> (&r)
   while\<^sup>\<top> (&r * &r \<le>\<^sub>u \<guillemotleft>a\<guillemotright>)
   do 
    assign_r r (&r + 1)
   od;;
   assign_r r (&r - 1)
 \<lbrace>0\<le>\<^sub>u &r \<and> uop power2 (&r) \<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> \<guillemotleft>a\<guillemotright> <\<^sub>u uop power2 (&r + 1)\<rbrace>\<^sub>u" 
  apply (insert assms)
  supply Isqrt_aux [simp]
  apply (vcg sp)    
  done 
    
lemma sqrt_prog_correct_wp_rel:
  assumes "vwb_lens r"
  shows
 "\<lbrace>0 \<le>\<^sub>u \<guillemotleft>a\<guillemotright>\<rbrace>
   assign_r r 1 ;; 
   invr 0\<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> bop Isqrt \<guillemotleft>a\<guillemotright> (&r)
   while\<^sup>\<top> (&r * &r \<le>\<^sub>u \<guillemotleft>a\<guillemotright>)
   do 
    assign_r r (&r + 1)
   od;;
   assign_r r (&r - 1)
 \<lbrace>0\<le>\<^sub>u &r \<and> uop power2 (&r) \<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> \<guillemotleft>a\<guillemotright> <\<^sub>u uop power2 (&r + 1)\<rbrace>\<^sub>u" 
  apply (insert assms)
  supply Isqrt_aux [simp]
  apply (vcg wp)    
  done  
     
subsection \<open>gcd\<close>
  
text \<open>In the following we illustrate the effect of domain theory based approach. 
      Namely, in the lemma gcd_correct we use the hard coded max function 
      @{term "(trop If (&r >\<^sub>u &x) (&r) (&x))"}. This leads to long proof. 
      However in gcd_correct' we use the max function from HOL library. 
      This leads to a shorter proof since max library contains the necessary lemmas that simplify
      the reasoning.\<close>
  
lemma gcd_correct_sp_H1_H3:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows  
 "\<lbrace> \<guillemotleft>b\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0 \<rbrace> 
   r :== \<guillemotleft>a\<guillemotright>;
   x :== \<guillemotleft>b\<guillemotright>;
   INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT (trop If (uop nat (&r) >\<^sub>u uop nat (&x)) (uop nat (&r)) (uop nat (&x)))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD
 \<lbrace> &r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright> \<rbrace>\<^sub>P"
  apply (insert assms)   
  apply (vcg sp)
     apply (auto simp: gcd_diff1_nat gcd_diff1)
  apply (metis gcd.commute gcd_diff1) +
  done  
     
lemma gcd_correct'_sp_H1_H3:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows  
 "\<lbrace>\<guillemotleft>b\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright> >\<^sub>u 0\<rbrace>
   r :== \<guillemotleft>a\<guillemotright>;
   x :== \<guillemotleft>b\<guillemotright>; 
   INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT  uop nat (bop max (&r) (&x))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD
 \<lbrace>&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>\<rbrace>\<^sub>P"
  apply (insert assms)  
  apply (vcg sp) 
   apply (simp add: gcd_diff1)
   apply (metis gcd.commute gcd_diff1)
  done
    
lemma gcd_correct'_wp_H1_H3:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows
 "\<lbrace>\<guillemotleft>b\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0\<rbrace>
   r :== \<guillemotleft>a\<guillemotright>;
   x :== \<guillemotleft>b\<guillemotright>; 
   INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT uop nat (bop max (&r) (&x))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD
 \<lbrace>&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>\<rbrace>\<^sub>P"
  apply (insert assms)      
  apply (vcg wp) 
      apply (simp_all add: gcd_diff1)
   apply (metis gcd.commute gcd_diff1)+
  done
      
lemma gcd_correct'_sp_rel:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows  
 "\<lbrace>\<guillemotleft>b\<guillemotright>>\<^sub>u 0 \<and> \<guillemotleft>a::nat\<guillemotright>>\<^sub>u 0\<rbrace>
   r :== \<guillemotleft>a\<guillemotright>;;
   x :== \<guillemotleft>b\<guillemotright>;;
   invr &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   while\<^sup>\<top> \<not>(&r =\<^sub>u &x)
   do
    bif &r >\<^sub>u &x
    then r :== (&r - &x) 
    else x :== (&x - &r) 
    eif
   od
 \<lbrace>&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>\<rbrace>\<^sub>u"
  apply (insert assms)
    apply (vcg sp)
  apply (simp add: gcd_diff1_nat)
  apply (metis gcd.commute gcd_diff1_nat not_le)
  done      

lemma gcd_correct'_wp_rel:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows  
 "\<lbrace>\<guillemotleft>b\<guillemotright>>\<^sub>u 0 \<and> \<guillemotleft>a::nat\<guillemotright>>\<^sub>u 0\<rbrace>
   r :== \<guillemotleft>a\<guillemotright>;;
   x :== \<guillemotleft>b\<guillemotright>;; 
   invr &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   while\<^sup>\<top> \<not>(&r =\<^sub>u &x)
   do
    bif &r >\<^sub>u &x
    then r :== (&r - &x) 
    else x :== (&x - &r) 
    eif
   od
 \<lbrace>&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>\<rbrace>\<^sub>u"
  apply (insert assms)  
  apply (vcg wp)    
 apply (simp_all add: gcd_diff1_nat)
    apply (metis gcd.commute gcd_diff1_nat not_le)    
   apply (metis diff_is_0_eq gcd.commute gcd_diff1_nat not_le_minus)
  done 
    
section \<open>Arrays\<close>
  
subsection \<open>Array Max program: one-variable loop\<close>

lemma max_program_correct_sp_H1_H3:
  assumes "r \<bowtie> i" 
  assumes "vwb_lens i" "vwb_lens r" 
  shows  
  "\<lbrace>uop length \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u1 \<and> &i =\<^sub>u 1 \<and> &r =\<^sub>u bop nth \<guillemotleft>a:: int list\<guillemotright> 0\<rbrace> 
     INVAR  0 <\<^sub>u &i \<and> &i \<le>\<^sub>u uop length \<guillemotleft>a\<guillemotright> \<and> &r =\<^sub>u uop Max (uop set (bop take (&i) \<guillemotleft>a\<guillemotright>)) 
     VRT (uop length \<guillemotleft>a\<guillemotright> - (&i))  
     WHILE \<not>(&i =\<^sub>u uop length \<guillemotleft>a\<guillemotright>) 
     DO 
        IF &r <\<^sub>u  bop nth \<guillemotleft>a\<guillemotright> (&i) 
        THEN r :== bop nth \<guillemotleft>a\<guillemotright> (&i)
        ELSE SKIP
        FI;
        i :== (&i + 1)
     OD   
   \<lbrace>&r =\<^sub>uuop Max (uop set \<guillemotleft>a\<guillemotright>)\<rbrace>\<^sub>P"  
  apply (insert assms)
  apply (vcg sp)        
  subgoal for _ 
    by (cases a; auto)
  subgoal for _ i                  
    apply (clarsimp simp: take_Suc_conv_app_nth)
    apply (subst Max_insert) by auto
  subgoal for _ i 
    apply (clarsimp simp: take_Suc_conv_app_nth)  
    apply (subst Max_insert) by auto
  done  

lemma max_program_correct_wp_H1_H3:
  assumes "r \<bowtie> i" 
  assumes "vwb_lens i" "vwb_lens r" 
  shows  
"\<lbrace>uop length \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u1 \<and> &i =\<^sub>u 1 \<and> &r =\<^sub>u bop nth \<guillemotleft>a:: int list\<guillemotright> 0\<rbrace> 
   INVAR  0 <\<^sub>u &i \<and> &i \<le>\<^sub>u uop length \<guillemotleft>a\<guillemotright> \<and> &r =\<^sub>u uop Max (uop set (bop take (&i) \<guillemotleft>a\<guillemotright>)) 
   VRT (uop length \<guillemotleft>a\<guillemotright> - (&i))  
   WHILE \<not>(&i =\<^sub>u uop length \<guillemotleft>a\<guillemotright>) 
   DO 
      IF &r <\<^sub>u  bop nth \<guillemotleft>a\<guillemotright> (&i) 
      THEN r :== bop nth \<guillemotleft>a\<guillemotright> (&i)
      ELSE SKIP
      FI;
      i :== (&i + 1)
   OD   
 \<lbrace>&r =\<^sub>uuop Max (uop set \<guillemotleft>a\<guillemotright>)\<rbrace>\<^sub>P"  
  apply (insert assms)
  apply (vcg wp)    
  subgoal for _ 
    by (cases a; auto)
  subgoal for _ i'
     apply (simp add: take_Suc_conv_app_nth )
    apply (subst (asm) Max_insert) 
      apply auto
    done
  subgoal for _ i' 
    apply (clarsimp simp: take_Suc_conv_app_nth)  
    apply (cases a, auto)
    done
  subgoal for _ i 
    by (clarsimp simp: take_Suc_conv_app_nth)  
  subgoal for _ i   
    by (clarsimp simp: take_Suc_conv_app_nth)  
  done  
    
lemma max_program_correct_sp_rel:
  assumes "r \<bowtie> i" 
  assumes "vwb_lens i" "vwb_lens r" 
  shows  
  "\<lbrace>uop length \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u1 \<and> &i =\<^sub>u 1 \<and> &r =\<^sub>u bop nth \<guillemotleft>a:: int list\<guillemotright> 0\<rbrace> 
     invr  0 <\<^sub>u &i \<and> &i \<le>\<^sub>u uop length \<guillemotleft>a\<guillemotright> \<and> &r =\<^sub>u uop Max (uop set (bop take (&i) \<guillemotleft>a\<guillemotright>)) 
     while\<^sup>\<top> \<not>(&i =\<^sub>u uop length \<guillemotleft>a\<guillemotright>) 
     do 
        bif &r <\<^sub>u  bop nth \<guillemotleft>a\<guillemotright> (&i) 
        then assign_r r (bop nth \<guillemotleft>a\<guillemotright> (&i))
        else SKIP\<^sub>r
        eif;;
        assign_r i (&i + 1)
     od   
   \<lbrace>&r =\<^sub>uuop Max (uop set \<guillemotleft>a\<guillemotright>)\<rbrace>\<^sub>u"  
  apply (insert assms)
  apply (vcg sp)        
  subgoal for _ 
    by (cases a; auto)
  subgoal for _ i                  
    apply (clarsimp simp: take_Suc_conv_app_nth)
    apply (subst Max_insert) by auto
  subgoal for _ i 
    apply (clarsimp simp: take_Suc_conv_app_nth)  
    apply (subst Max_insert) by auto
  done 

lemma max_program_correct_wp_rel:
  assumes "r \<bowtie> i" 
  assumes "vwb_lens i" "vwb_lens r" 
  shows  
  "\<lbrace>uop length \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u\<guillemotleft>1\<guillemotright> \<and> &i =\<^sub>u \<guillemotleft>1\<guillemotright> \<and> &r =\<^sub>u bop nth \<guillemotleft>a:: int list\<guillemotright> \<guillemotleft>0\<guillemotright>\<rbrace> 
     invr  0 <\<^sub>u &i \<and> &i \<le>\<^sub>u uop length \<guillemotleft>a\<guillemotright> \<and> &r =\<^sub>u uop Max (uop set (bop take (&i) \<guillemotleft>a\<guillemotright>)) 
     while\<^sup>\<top> \<not>(&i =\<^sub>u uop length \<guillemotleft>a\<guillemotright>) 
     do 
        bif &r <\<^sub>u  bop nth \<guillemotleft>a\<guillemotright> (&i) 
        then  r :==(bop nth \<guillemotleft>a\<guillemotright> (&i))
        else SKIP\<^sub>r
        eif;;
         i :==(&i + \<guillemotleft>1\<guillemotright>)
     od   
   \<lbrace>&r =\<^sub>uuop Max (uop set \<guillemotleft>a\<guillemotright>)\<rbrace>\<^sub>u"  
  apply (insert assms)
  apply (vcg wp)
  subgoal for _ 
    by (cases a; auto)
  subgoal for _ i'
    apply (simp add: take_Suc_conv_app_nth )
    apply (subst  Max_insert) 
      apply auto
    done
  subgoal for _ i' 
    apply (clarsimp simp: take_Suc_conv_app_nth) 
    apply (subst (asm) Max_insert) 
      apply auto
    done
  subgoal for _ i 
    apply (simp add: take_Suc_conv_app_nth )
    apply (subst (asm) Max_insert) 
      apply auto
    apply (subst  Max_insert) 
      apply auto
    done
  done

section \<open>Modular reasoning\<close>
abbreviation
  "swaping a j \<equiv> ((a :== (trop swap (\<guillemotleft>j\<guillemotright>) (\<guillemotleft>j\<guillemotright>-1) (&a)))::_ prog)"

lemma  swap_preserves_list:
  assumes "vwb_lens a"
  assumes "vwb_lens b"
  assumes "a \<bowtie> b"
  shows 
    "\<lbrace>\<lambda>aa. ((&a) =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and>\<guillemotleft>j\<guillemotright> <\<^sub>u uop length \<guillemotleft>aa\<guillemotright>)\<rbrace>
      \<lambda>aa. swaping a j;
           (b):== (\<guillemotleft>0\<guillemotright>)
     \<lbrace>\<lambda>aa. (mset\<^sub>u((&a)) =\<^sub>u mset\<^sub>u(\<guillemotleft>aa\<guillemotright>))\<rbrace>\<^sub>G"
  by (insert assms, rule allI,vcg wp)

(*we add 2 massaging of the same proved spec to the VCG since depending on the situation
  the VCG may use one or another. Namely, there is a unification problem that needs to be solved*)

declare swap_preserves_list[THEN spec, vcg_proved_specs]
declare swap_preserves_list[simplified, THEN spec, vcg_proved_specs]

lemma "vwb_lens x \<Longrightarrow> x \<subseteq>\<^sub>L  id_lens"

  by (simp add: sublens_greatest)
  
lemma swap_preserves_list_test:
  assumes "vwb_lens a"
  assumes "vwb_lens b"
  shows 
    "\<lbrace>\<lambda>aa. ((&a:fst_lens) =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and>\<guillemotleft>j\<guillemotright> <\<^sub>u uop length \<guillemotleft>aa\<guillemotright>)\<rbrace>
      \<lambda>aa. (&a:fst_lens):[swaping (a;\<^sub>L fst_lens) j;
           (b:snd_lens):== (\<guillemotleft>0\<guillemotright>)]
   \<lbrace>\<lambda>aa. (mset\<^sub>u(&a:fst_lens) =\<^sub>u mset\<^sub>u(\<guillemotleft>aa\<guillemotright>))\<rbrace>\<^sub>G"
  apply (insert assms)
  apply (vcg wp)
  apply (insert assms)  
  apply (subst seq_prog_anti_frame_prog_modifies)
     defer
     apply (subst assigns_prog_antiframe_prog_modifies)
      defer
      apply simp
     apply (subst assigns_prog_antiframe_prog_modifies)
      defer
      apply simp
     apply (simp add: DONT_TOUCH_def prog_rep_eq)
    apply simp
  unfolding  pr_var_def
   apply rel_simp+
  done

abbreviation
  "count_down j \<equiv> ((j :== ((&j) - \<guillemotleft>1\<guillemotright>))::_ prog)"

lemma count_down_preserves_list:
  assumes "vwb_lens j" 
  shows  
  "\<lbrace>\<lambda>n. (&j) =\<^sub>u \<guillemotleft>n::nat\<guillemotright> \<and> (&j) \<noteq>\<^sub>u \<guillemotleft>0::nat\<guillemotright>\<rbrace> 
     \<lambda>aa.  count_down j
   \<lbrace>\<lambda>n. (&j) \<le>\<^sub>u \<guillemotleft>n\<guillemotright>\<rbrace>\<^sub>G" 
  by (insert assms, rule allI, vcg wp)

declare count_down_preserves_list[ THEN spec, vcg_proved_specs]
declare count_down_preserves_list[simplified,THEN spec, vcg_proved_specs]

lemma count_down_preserves_list_test:
  assumes "vwb_lens j" 
  shows  
    "\<lbrace>\<lambda>n. (&j:fst_lens) =\<^sub>u \<guillemotleft>n::nat\<guillemotright> \<and> (&j:fst_lens) \<noteq>\<^sub>u \<guillemotleft>0::nat\<guillemotright>\<rbrace> 
      \<lambda>aa.  (&j:fst_lens):[count_down  (j;\<^sub>Lfst_lens)]
    \<lbrace>\<lambda>n. (&j:fst_lens) \<le>\<^sub>u \<guillemotleft>n\<guillemotright>\<rbrace>\<^sub>G" 
  apply (insert assms)
  apply (vcg wp)
    (*additional post-processing for proving VCs related to the modify assumption
       TODO: automate this*)
  apply (insert assms)
  apply (simp_all add: DONT_TOUCH_def prog_rep_eq)
  apply rel_auto+
  done

lemma swap_count_down_preserves_list:
  assumes "vwb_lens j" 
  assumes "vwb_lens a"
  assumes "vwb_lens b"
  assumes "(a:: 'b list \<Longrightarrow> 'a) \<bowtie> (j:: nat \<Longrightarrow> 'a)"
  shows  
    "\<lbrace>\<lambda>(aa,n). (&a:fst_lens) =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and> \<guillemotleft>i\<guillemotright> <\<^sub>u uop length \<guillemotleft>aa\<guillemotright> \<and> (&j:fst_lens) =\<^sub>u \<guillemotleft>n::nat\<guillemotright> \<and> \<guillemotleft>n\<guillemotright> \<noteq>\<^sub>u \<guillemotleft>0::nat\<guillemotright>\<rbrace> 
    \<lambda>aa. (&a:fst_lens):[swaping(a;\<^sub>L fst_lens) i;(b:snd_lens):== (\<guillemotleft>0\<guillemotright>)];
         (&j:fst_lens):[(j:fst_lens) :== ((&j:fst_lens) - \<guillemotleft>1\<guillemotright>)]
   \<lbrace>\<lambda>(aa,n).  (&j:fst_lens) \<le>\<^sub>u \<guillemotleft>n\<guillemotright>\<and> mset\<^sub>u((&a:fst_lens)) =\<^sub>u mset\<^sub>u(\<guillemotleft>aa\<guillemotright>)\<rbrace>\<^sub>G" 
  apply (insert assms)
  apply (vcg wp) (*Why wp march et wp_trace ne marche pas!!*)
    (*additional post-processing for proving VCs related to the modify assumption
       TODO: automate this*)
   apply (insert assms)
   apply (subst seq_prog_anti_frame_prog_modifies)
     defer
     apply (subst assigns_prog_antiframe_prog_modifies)
      defer
      apply simp
     apply (subst assigns_prog_antiframe_prog_modifies)
      defer
      apply simp
     apply (simp_all add: DONT_TOUCH_def prog_rep_eq)
   apply rel_simp+
  done


lemma blah:
  assumes "r \<bowtie> i" 
  assumes "vwb_lens i" "vwb_lens r" 
  shows  
"\<lbrace> (\<lambda>(aa).  (\<guillemotleft>a\<guillemotright> =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and> uop length \<guillemotleft>aa\<guillemotright> \<ge>\<^sub>u\<guillemotleft>1\<guillemotright> \<and> &i =\<^sub>u \<guillemotleft>1\<guillemotright> \<and> &r =\<^sub>u bop nth \<guillemotleft>aa:: int list\<guillemotright> \<guillemotleft>0\<guillemotright>))\<rbrace> 
   \<lambda>(aa). 
  INVAR  ( (\<guillemotleft>a\<guillemotright> =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and> \<guillemotleft>0\<guillemotright> <\<^sub>u &i \<and> &i \<le>\<^sub>u uop length \<guillemotleft>aa\<guillemotright> \<and> &r =\<^sub>u uop Max (uop set (bop take (&i) \<guillemotleft>aa\<guillemotright>)))) 
   VRT   (uop length \<guillemotleft>a\<guillemotright> - (&i))
   WHILE \<not>(&i =\<^sub>u uop length \<guillemotleft>a\<guillemotright>)
     DO IF &r <\<^sub>u  bop nth \<guillemotleft>a\<guillemotright> (&i) 
        THEN r :== bop nth \<guillemotleft>a\<guillemotright> (&i)
        ELSE SKIP
        FI;
        i :== (&i + \<guillemotleft>1\<guillemotright>)
     OD
 \<lbrace> (\<lambda>(aa).  (&r =\<^sub>uuop Max (uop set \<guillemotleft>aa\<guillemotright>)))\<rbrace>\<^sub>G"  
  apply (insert assms) 
  apply (vcg wp)
  subgoal for _ 
    by (cases a; auto)
  subgoal for _ i'
     apply (simp add: take_Suc_conv_app_nth )
    apply (subst (asm) Max_insert) 
      apply auto
    done
  subgoal for _ i' 
    apply (clarsimp simp: take_Suc_conv_app_nth)  
    apply (cases a, auto)
    done
  subgoal for _ i 
    by (clarsimp simp: take_Suc_conv_app_nth)  
  subgoal for _ i   
    by (clarsimp simp: take_Suc_conv_app_nth)  
  done 

(*TODO: CONTINUE FROM HERE!!*)

declare blah[unfolded Product_Type.split_paired_All, unfolded Product_Type.prod.case,THEN spec, vcg_proved_specs]

lemma blah':
  assumes "r \<bowtie> i" 
  assumes "vwb_lens i" "vwb_lens r" 
  shows  
"\<lbrace> (\<lambda>(aa).  (\<guillemotleft>a\<guillemotright> =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and> uop length \<guillemotleft>aa\<guillemotright> \<ge>\<^sub>u\<guillemotleft>1\<guillemotright> \<and> (&i:fst_lens) =\<^sub>u \<guillemotleft>1\<guillemotright> \<and> 
            (&r:fst_lens) =\<^sub>u bop nth \<guillemotleft>aa:: int list\<guillemotright> \<guillemotleft>0\<guillemotright>))\<rbrace> 
   \<lambda>(aa). 
  IN  ((i;\<^sub>Lfst_lens) +\<^sub>L (r;\<^sub>Lfst_lens))
  INVAR  ((\<guillemotleft>a\<guillemotright> =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and> \<guillemotleft>0\<guillemotright> <\<^sub>u (&i:fst_lens) \<and> (&i:fst_lens) \<le>\<^sub>u uop length \<guillemotleft>aa\<guillemotright> \<and> 
           (&r:fst_lens) =\<^sub>u uop Max (uop set (bop take (&i:fst_lens) \<guillemotleft>aa\<guillemotright>)))) 
   VRT     (uop length \<guillemotleft>a\<guillemotright> - (&i:fst_lens))
   WHILE \<not>((&i:fst_lens) =\<^sub>u uop length \<guillemotleft>a\<guillemotright>)
     DO IF (&r:fst_lens) <\<^sub>u  bop nth \<guillemotleft>a\<guillemotright> (&i:fst_lens) 
        THEN (r:fst_lens) :== (bop nth \<guillemotleft>a\<guillemotright> (&i:fst_lens))
        ELSE SKIP
        FI;
        (i:fst_lens) :== ((&i:fst_lens) + \<guillemotleft>1\<guillemotright>)
     OD
 \<lbrace> (\<lambda>(aa).  ((&r:fst_lens) =\<^sub>uuop Max (uop set \<guillemotleft>aa\<guillemotright>)))\<rbrace>\<^sub>G"  
  apply (insert assms) 
  apply (vcg wp)
  subgoal for _ 
    by (cases a; auto)
        defer
        defer
        defer
  subgoal for _ i'
     apply (simp add: take_Suc_conv_app_nth )
    apply (subst (asm) Max_insert) 
      apply auto
    done
  subgoal for _ i' 
    apply (clarsimp simp: take_Suc_conv_app_nth)  
    apply (cases a, auto)
    done
  subgoal for _ i 
    by (clarsimp simp: take_Suc_conv_app_nth)  
  subgoal for _ i   
    by (clarsimp simp: take_Suc_conv_app_nth)  
    apply (simp add: assms(1) assms(2) assms(3)  lens_indep_sym plus_vwb_lens)
  apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_pred_sublens)
  apply (subst seq_prog_anti_frame_prog_modifies)
  defer
  apply (subst If_prog_anti_frame_prog_modifies )
       apply (subst assigns_prog_antiframe_prog_modifies)
        defer
        defer
        apply (subst SKIP_prog_anti_frame_prog_modifies)
        apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
       apply simp
      apply simp
       apply (subst assigns_prog_antiframe_prog_modifies)
  using assms
       apply rel_simp
        apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
     apply simp
     apply (simp add: DONT_TOUCH_def)
  apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym)
  using assms
   apply rel_simp
  apply simp
  done

declare blah'[unfolded Product_Type.split_paired_All, unfolded Product_Type.prod.case,THEN spec, vcg_proved_specs]

lemma
  assumes "r \<bowtie> i" 
  assumes "vwb_lens i" "vwb_lens r" 
  shows  
"\<lbrace>\<lambda>(aa). \<guillemotleft>a\<guillemotright> =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and> uop length \<guillemotleft>aa\<guillemotright> \<ge>\<^sub>u\<guillemotleft>1\<guillemotright> \<and> (&i:fst_lens) =\<^sub>u \<guillemotleft>1\<guillemotright> \<and> 
         (&r:fst_lens) =\<^sub>u bop nth \<guillemotleft>aa:: int list\<guillemotright> \<guillemotleft>0\<guillemotright>\<rbrace> 
  \<lambda>(aa).  MODIFY i ;\<^sub>L fst_lens +\<^sub>L r ;\<^sub>Lfst_lens DO
   IN ((i;\<^sub>Lfst_lens) +\<^sub>L (r;\<^sub>Lfst_lens)) 
   INVAR (\<guillemotleft>a\<guillemotright> =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and> \<guillemotleft>0\<guillemotright> <\<^sub>u (&i:fst_lens) \<and> (&i:fst_lens) \<le>\<^sub>u uop length \<guillemotleft>aa\<guillemotright> \<and> 
          (&r:fst_lens) =\<^sub>u uop Max (uop set (bop take (&i:fst_lens) \<guillemotleft>aa\<guillemotright>)))
   VRT(uop length \<guillemotleft>a\<guillemotright> - (&i:fst_lens))
   WHILE \<not>((&i:fst_lens) =\<^sub>u uop length \<guillemotleft>a\<guillemotright>)
       DO IF (&r:fst_lens) <\<^sub>u  bop nth \<guillemotleft>a\<guillemotright> (&i:fst_lens) 
          THEN (r:fst_lens) :== (bop nth \<guillemotleft>a\<guillemotright> (&i:fst_lens))
          ELSE SKIP
          FI;
          (i:fst_lens) :== ((&i:fst_lens) + \<guillemotleft>1\<guillemotright>)
       OD 
      OD
 \<lbrace>\<lambda>(aa). (&r:fst_lens) =\<^sub>uuop Max (uop set \<guillemotleft>aa\<guillemotright>)\<rbrace>\<^sub>G"  
  apply (insert assms) 
  apply (vcg wp)
    apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
   apply (metis assms(1) assms(2) assms(3) fst_vwb_lens in_var_def in_var_indep lens_comp_lb lens_indep_sym plus_pred_sublens vwb_lens_def)
 
  unfolding  DONT_TOUCH_def while_lfp_scp_invr_vrt_prog_def
  thm modifies_indep
  apply (subst modifies_indep)
   (**)
  apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
  apply simp
  apply simp
  defer
   apply blast
  (**)
  unfolding  pr_var_def thm antiframe_prog_modifies_while_lfp_scp_prog
  apply (subst antiframe_prog_modifies_while_lfp_scp_prog)
  (**)
  apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
  defer
  apply blast
  (**)
   apply (subst seq_prog_anti_frame_prog_modifies)
  defer
  apply (subst If_prog_anti_frame_prog_modifies )
       apply (subst assigns_prog_antiframe_prog_modifies)
        defer
        defer
        apply (subst SKIP_prog_anti_frame_prog_modifies)
        apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
       apply simp
      apply simp
       apply (subst assigns_prog_antiframe_prog_modifies)
  using assms
       apply rel_simp
        apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
     apply simp
    apply (simp add: DONT_TOUCH_def)
  apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym)
  using assms
   apply rel_simp
  apply simp
  done
abbreviation gvar :: "('a \<Longrightarrow> 'b) \<Rightarrow> 'a \<Longrightarrow> 'b \<times> 'c" ("GVAR (_)/" [1000])
  where  "gvar x \<equiv> x ;\<^sub>L fst_lens"

abbreviation lvar :: "('a \<Longrightarrow> 'b) \<Rightarrow> 'a \<Longrightarrow>'c \<times>'b " ("LVAR (_)/" [1000])
  where  "lvar x \<equiv> x ;\<^sub>L snd_lens"

term "MODIFY GVAR x +\<^sub>L GVAR y DO body OD"
term "MODIFY ((i;\<^sub>Lfst_lens) +\<^sub>L (r;\<^sub>Lfst_lens)) DO body OD"

lemma
  assumes "r \<bowtie> i" 
  assumes "vwb_lens i" "vwb_lens r" 
  shows  
"\<lbrace>\<lambda>aa. \<guillemotleft>a\<guillemotright> =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and> uop length \<guillemotleft>aa\<guillemotright> \<ge>\<^sub>u\<guillemotleft>1\<guillemotright> \<and> (&i:fst_lens) =\<^sub>u \<guillemotleft>1\<guillemotright> \<and> 
         (&r:fst_lens) =\<^sub>u bop nth \<guillemotleft>aa:: int list\<guillemotright> \<guillemotleft>0\<guillemotright>\<rbrace> 
  \<lambda>aa.  {&i:fst_lens, &r:fst_lens}:[
   IN ((i;\<^sub>Lfst_lens) +\<^sub>L (r;\<^sub>Lfst_lens)) 
   INVAR (\<guillemotleft>a\<guillemotright> =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and> \<guillemotleft>0\<guillemotright> <\<^sub>u (&i:fst_lens) \<and> (&i:fst_lens) \<le>\<^sub>u uop length \<guillemotleft>aa\<guillemotright> \<and> 
          (&r:fst_lens) =\<^sub>u uop Max (uop set (bop take (&i:fst_lens) \<guillemotleft>aa\<guillemotright>)))
   VRT(uop length \<guillemotleft>a\<guillemotright> - (&i:fst_lens))
   WHILE \<not>((&i:fst_lens) =\<^sub>u uop length \<guillemotleft>a\<guillemotright>)
       DO IF (&r:fst_lens) <\<^sub>u  bop nth \<guillemotleft>a\<guillemotright> (&i:fst_lens) 
          THEN (r:fst_lens) :== (bop nth \<guillemotleft>a\<guillemotright> (&i:fst_lens))
          ELSE SKIP
          FI;
          (i:fst_lens) :== ((&i:fst_lens) + \<guillemotleft>1\<guillemotright>)
       OD]
 \<lbrace>\<lambda>(aa). (&r:fst_lens) =\<^sub>uuop Max (uop set \<guillemotleft>aa\<guillemotright>)\<rbrace>\<^sub>G"  
  apply (insert assms) 
  apply (vcg wp)
    apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
   apply (metis assms(1) assms(2) assms(3) fst_vwb_lens in_var_def in_var_indep lens_comp_lb lens_indep_sym plus_pred_sublens vwb_lens_def) 
  (*apply ( ((match conclusion in "DONT_TOUCH (g:[IN g INVAR _ VRT _WHILE _ DO body OD] = IN g INVAR _ VRT _ WHILE _ DO body OD)" for g body \<Rightarrow> succeed), 
            (simp only: while_lfp_scp_invr_vrt_prog_def DONT_TOUCH_def pr_var_def), subst antiframe_prog_modifies_while_lfp_scp_prog)
        | ((match conclusion in "DONT_TOUCH (g:[IN g INVAR _ WHILE _ DO body OD] = IN g INVAR _ WHILE _ DO body OD)" for g body \<Rightarrow> succeed), 
            (simp only: while_lfp_scp_invr_prog_def DONT_TOUCH_def pr_var_def), subst antiframe_prog_modifies_while_lfp_scp_prog)
        | ((match conclusion in "DONT_TOUCH (g:[IN g WHILE _ DO body OD] = IN g WHILE _ DO body OD)" for g body \<Rightarrow> succeed), 
            (simp only: DONT_TOUCH_def pr_var_def), subst antiframe_prog_modifies_while_lfp_scp_prog)
        | ((match conclusion in "DONT_TOUCH (g':[IN g INVAR _ VRT _WHILE _ DO body OD] = IN g INVAR _ VRT _ WHILE _ DO body OD)" for g g' body \<Rightarrow> succeed), 
            (simp only: while_lfp_scp_invr_vrt_prog_def DONT_TOUCH_def), subst modifies_indep)
        | ((match conclusion in "DONT_TOUCH (g':[IN g INVAR _ WHILE _ DO body OD] = IN g INVAR _ WHILE _ DO body OD)" for g g' body \<Rightarrow> succeed), 
            (simp only: while_lfp_scp_invr_prog_def DONT_TOUCH_def), subst modifies_indep(*This rule is very dangerous if not applied correctly*))
        | ((match conclusion in "DONT_TOUCH (g':[IN g WHILE _ DO body OD] = IN g WHILE _ DO body OD)" for g g' body \<Rightarrow> succeed), 
            (simp only: DONT_TOUCH_def), subst modifies_indep)
        | ((match conclusion in "_" \<Rightarrow> succeed), vcg_defer)
          (*TODO: add the other loop patterns*)
         )*)
  unfolding DONT_TOUCH_def while_lfp_scp_invr_vrt_prog_def
 apply (subst modifies_indep (**))
   (**)
  apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
  apply simp
  apply simp
  defer
   apply blast
  (**)
  unfolding  pr_var_def
  apply (subst antiframe_prog_modifies_while_lfp_scp_prog)
  (**)
  apply (simp add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
  defer
  apply blast
  (**)
   apply (subst seq_prog_anti_frame_prog_modifies)
  defer
     apply (subst If_prog_anti_frame_prog_modifies )
       apply (subst assigns_prog_antiframe_prog_modifies)   
        defer
        defer
        apply (subst SKIP_prog_anti_frame_prog_modifies)
         defer
         defer
         defer
         apply (subst assigns_prog_antiframe_prog_modifies)
  using assms
        apply (simp_all add: assms(1) assms(2) assms(3) lens_indep_sym plus_vwb_lens)
  using assms
       apply rel_simp+
  done
(* 


TODO List for next iteration:
- Implement a debuging version of SP
- renovate the old VCG normal mode
- implement VCG proof_annotation_mode
*) 

find_theorems name: "rep_eq" "(Rep_uexpr ?e = ?t)" (*This what pred_simp uses...*)
    
(*lemma demo_VAR_BIND:
  assumes " r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  assumes VAR_BIND: "(get\<^bsub>r\<^esub> )  = r\<^sub>0 \<and> (get\<^bsub>x\<^esub> )  = x\<^sub>0"  
  shows  
"\<lbrace>uop length \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u1 \<and> &i =\<^sub>u 1 \<and> &r =\<^sub>u bop nth \<guillemotleft>a:: int list\<guillemotright> 0\<rbrace> 
   INVAR  0 <\<^sub>u &i \<and> &i \<le>\<^sub>u uop length \<guillemotleft>a\<guillemotright> \<and> &r =\<^sub>u uop Max (uop set (bop take (&i) \<guillemotleft>a\<guillemotright>)) 
   VRT (uop length \<guillemotleft>a\<guillemotright> - (&i))  
   WHILE \<not>(&i =\<^sub>u uop length \<guillemotleft>a\<guillemotright>) 
   DO 
      IF &r <\<^sub>u  bop nth \<guillemotleft>a\<guillemotright> (&i) 
      THEN r :== bop nth \<guillemotleft>a\<guillemotright> (&i)
      ELSE SKIP
      FI;
      i :== (&i + 1)
   OD
 \<lbrace>&r =\<^sub>uuop Max (uop set \<guillemotleft>a\<guillemotright>)\<rbrace>\<^sub>P"
  apply (insert assms(1) assms(2))  
  apply(vcg sp)
    apply (auto simp only: VAR_BIND assms(3))
   apply (simp add: gcd_diff1_nat)
  apply (metis gcd.commute gcd_diff1_nat not_le)
  done
*)
section\<open>DEMO for York Talk\<close>

lemma EXAMPLE:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows
 "\<lbrace>&r =\<^sub>u \<guillemotleft>a\<guillemotright> \<and> &x =\<^sub>u \<guillemotleft>b\<guillemotright> \<and> \<guillemotleft>b\<guillemotright>>\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0\<rbrace> 
   INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT uop nat (bop max (&r) (&x))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD
 \<lbrace>&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>\<rbrace>\<^sub>P"
  apply (insert assms)      
  apply (vcg wp) 
      apply (simp_all add: gcd_diff1)
   apply (metis gcd.commute gcd_diff1)+
  done

lemma EXAMPLE_WITHOUT_Backend:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows
 "\<lbrace>&r =\<^sub>u \<guillemotleft>a\<guillemotright> \<and> &x =\<^sub>u \<guillemotleft>b\<guillemotright> \<and> \<guillemotleft>b\<guillemotright>>\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0\<rbrace> 
   INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT uop nat (bop max (&r) (&x))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD
 \<lbrace>&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>\<rbrace>\<^sub>P"
  apply (insert assms)      
  apply (wp) 
(*sledgehammer*)
     apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  (*sledgehammer*)
      apply (simp_all add: gcd_diff1)
   apply (metis gcd.commute gcd_diff1)+
  done

term "SKIP"

term "(x :== e) :: '\<alpha> prog"

term "IF b THEN P\<^sub>1 ELSE P\<^sub>2 FI"

term "WHILE b DO body OD"

term "INVAR invar VRT vari WHILE b DO body OD"

term "FROM init UNTIL exit DO body OD"

term "FOR (init, b, incr) DO body OD"

term "DO body WHILE b OD"

term "MODIFY VARS DO body OD"

term "(X:[P]  = P)"

term "frame X (assign_r (x ) v)"
term "(antiframe X (assign_r (x;\<^sub>L X) v)  = (assign_r (x;\<^sub>L X) v))"

lemma \<open>mwb_lens X \<Longrightarrow> mwb_lens x \<Longrightarrow> x \<subseteq>\<^sub>L X \<Longrightarrow>
       antiframe X (assign_r (x) v)  = (assign_r (x) v)\<close>
  by rel_auto

lemma antiframe_prog_hoare_basic_weak:
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_g':\<open>vwb_lens g'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>g \<bowtie> l\<close> 
  assumes region     :\<open>g' \<subseteq>\<^sub>L g\<close>
  assumes modifies   :\<open>{&g', &l}:[PROG] = PROG\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>  
  shows \<open>\<lbrace>P \<and> R\<rbrace>MODIFY g' DO PROG OD\<lbrace>(\<exists> {&l, &g'} \<bullet> Q \<oplus>\<^sub>p g') \<and>  (\<exists> {&l, &g'} \<bullet> R \<oplus>\<^sub>p l)\<rbrace>\<^sub>P\<close>
  using vwb_lens_g vwb_lens_g' vwb_lens_l indep_g_l modifies hoare  
  unfolding lens_indep_def
  apply (simp add:prog_rep_eq)
  apply rel_simp
  using indep_g_l 
  apply rel_simp
  apply auto
   apply(insert region)
  subgoal for  _ more _ x
    apply (rule exI[where x = \<open>get\<^bsub>l\<^esub>x\<close>])
    using region unfolding lens_indep_def
    apply auto
    apply (drule  sublens_pres_indep[OF _ indep_g_l])
    unfolding lens_indep_def
    apply simp apply blast
    done
  done

term "Rep_uexpr(\<exists>g' \<bullet> R)"
lemma \<open>\<lbrakk>\<exists> g' \<bullet> R\<rbrakk>\<^sub>e s = (\<exists>v. \<lbrakk>R\<rbrakk>\<^sub>e (put\<^bsub>g\<^esub> s v))\<close>
  apply rel_auto
  oops

lemma
  assumes vwb_lens_g :\<open>vwb_lens g\<close> 
  assumes vwb_lens_g':\<open>vwb_lens g'\<close> 
  assumes vwb_lens_l :\<open>vwb_lens l\<close>
  assumes indep_g_l  :\<open>g \<bowtie> l\<close> 
  assumes region     :\<open>g' \<subseteq>\<^sub>L g\<close>
  assumes strict     :\<open>(g +\<^sub>L l) = 1\<^sub>L\<close>  
  assumes modifies   :\<open>{&g', &l}:[PROG] = PROG\<close> (*TODO: From algebraic laws of program of antiframe this assumption can be derived from {&g'}:[PROG] = PROG*)
  assumes hoare      :\<open>\<lbrace>P\<rbrace>PROG\<lbrace>Q\<rbrace>\<^sub>P\<close>  
  shows \<open>\<lbrace>P \<and> R\<rbrace>MODIFY g' DO PROG OD\<lbrace>(\<exists>l \<bullet> Q ) \<and>  (\<exists>g' \<bullet> R)\<rbrace>\<^sub>P\<close>
  using vwb_lens_g vwb_lens_g' vwb_lens_l indep_g_l modifies hoare   strict
  unfolding lens_indep_def
  apply (simp add:prog_rep_eq)
  apply rel_simp
  using indep_g_l 
  apply rel_simp
  apply auto
  apply(insert region)
  subgoal for  _ more more' _ _ _ ab bb
    apply (rule exI[where x = \<open>get\<^bsub>l\<^esub>(ab , bb )\<close>])
    using region unfolding lens_indep_def
    apply auto
    apply (drule  sublens_pres_indep[OF _ indep_g_l])
    unfolding lens_indep_def
    apply simp  
    apply (metis mwb_lens_def vwb_lens_mwb weak_lens.put_get)
    done
  subgoal for _ a b _ _ _ ab bb
    apply (rule exI[where x = \<open>get\<^bsub>g'\<^esub>(a , b)\<close>])
    unfolding lens_indep_def
    apply (metis sublens_put_put vwb_lens_mwb vwb_lens_wb wb_lens.get_put)
    done
  done


lemma \<open>&x = utp_expr.var (x )\<close>
  apply rel_simp
  done

lemma \<open>\<forall> s s'. Rep_uexpr ($x) (s, s') = Rep_uexpr (utp_expr.var (x ;\<^sub>L fst_lens)) (s, s')\<close>
 apply rel_simp
done

lemma  \<open>bop f x y = sda\<close>
  apply rel_simp
  oops
lemma \<open>(\<lambda> x \<bullet> e) = sda\<close>
  apply rel_simp
  oops

  term "subst_upd_uvar \<sigma> X e"
  thm swap_usubst_inj
  term \<open>subst (subst_upd_uvar id X v) e\<close>
end

