(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)

theory algorithms_presentation

imports "../Front-End/Syntax_Interface" 
        "../Midend-IVL/Isabelle-UTP-Extended/utils/utp_extensions" 

begin
section \<open>setup and makeup!\<close>    

abbreviation gvar :: "('a \<Longrightarrow> 'b) \<Rightarrow> 'a \<Longrightarrow> 'b \<times> 'c" ("GVAR (_)/" [1000])
  where  "gvar x \<equiv> x ;\<^sub>L fst_lens"

abbreviation lvar :: "('a \<Longrightarrow> 'b) \<Rightarrow> 'a \<Longrightarrow>'c \<times>'b " ("LVAR (_)/" [1000])
  where  "lvar x \<equiv> x ;\<^sub>L snd_lens"

sledgehammer_params[stop_on_first,parallel_subgoals, join_subgoals]

no_adhoc_overloading'_all

section \<open>Simple algorithms\<close>
subsection \<open>add program\<close>

program_spec add_prog_sp_total:
  assumes_utp "(\<guillemotleft>y::nat\<guillemotright>) >\<^sub>u 0"
  ensures_utp "&a =\<^sub>u \<guillemotleft>y\<guillemotright> + \<guillemotleft>x\<guillemotright>"
  prog_utp  
  "a :== \<guillemotleft>x\<guillemotright> ;
   b :== \<guillemotleft>y\<guillemotright>;
   INVAR  &a =\<^sub>u  \<guillemotleft>x\<guillemotright> + (\<guillemotleft>y\<guillemotright> - &b)  \<and>  \<guillemotleft>y\<guillemotright> \<ge>\<^sub>u &b 
   VRT (&b)
   WHILE &b \<noteq>\<^sub>u \<guillemotleft>0\<guillemotright>  
     DO a:== (&a + \<guillemotleft>1\<guillemotright>);
        b :== (&b - \<guillemotleft>1\<guillemotright>)
     OD"
  vcg_sp
  done

program_spec add_prog_wp_total:
  assumes_utp "(\<guillemotleft>y::nat\<guillemotright>) >\<^sub>u 0"
  ensures_utp "&a =\<^sub>u \<guillemotleft>y\<guillemotright> + \<guillemotleft>x\<guillemotright>"
  prog_utp  
  "a :== \<guillemotleft>x\<guillemotright> ;
   b :== \<guillemotleft>y\<guillemotright>;
   INVAR  &a =\<^sub>u  \<guillemotleft>x\<guillemotright> + (\<guillemotleft>y\<guillemotright> - &b)  \<and>  \<guillemotleft>y\<guillemotright> \<ge>\<^sub>u &b 
   VRT (&b)
   WHILE &b \<noteq>\<^sub>u \<guillemotleft>0\<guillemotright>  
     DO a:== (&a + \<guillemotleft>1\<guillemotright>);
        b :== (&b - \<guillemotleft>1\<guillemotright>)
     OD"
  vcg_wp
  done

subsection \<open>Increment method\<close>
  
program_spec increment_method_sp_H1_H3:
  assumes_utp "\<guillemotleft>a\<guillemotright> >\<^sub>u 0"
  ensures_utp "\<guillemotleft>a\<guillemotright> =\<^sub>u &x"
  prog_utp    "x :== 0 ; 
               INVAR \<guillemotleft>a\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u &x 
               VRT (uop nat (\<guillemotleft>a\<guillemotright> - &x)) 
               WHILE &x <\<^sub>u \<guillemotleft>a\<guillemotright> DO x:== (&x + 1) OD"
  vcg_sp
  done

program_spec increment_method_wp_H1_H3:
  assumes_utp "\<guillemotleft>a\<guillemotright> >\<^sub>u 0"
  prog_utp    "x :== 0 ; 
               INVAR \<guillemotleft>a\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u &x 
               VRT (uop nat (\<guillemotleft>a\<guillemotright> - &x)) 
               WHILE &x <\<^sub>u \<guillemotleft>a\<guillemotright> DO x:== (&x + 1) OD"
  ensures_utp "\<guillemotleft>a\<guillemotright> =\<^sub>u &x"
  vcg_wp
  done
    
subsection \<open>even count program\<close> 

program_spec even_count_gen_sp_H1_H3:
  assumes_utp "\<guillemotleft>a\<guillemotright> >\<^sub>u 0"
  prog_utp    "i :== \<guillemotleft>0::int\<guillemotright>;
               j :== 0 ; 
               INVAR  (&j =\<^sub>u (&i + 1) div \<guillemotleft>2\<guillemotright> \<and> &i \<le>\<^sub>u \<guillemotleft>a\<guillemotright>) 
               VRT (uop nat (\<guillemotleft>a\<guillemotright> - &i))
               WHILE &i <\<^sub>u \<guillemotleft>a\<guillemotright>
               DO
                 IF &i mod \<guillemotleft>2\<guillemotright> =\<^sub>u 0 
                 THEN j :== (&j + 1)
                 ELSE SKIP 
                 FI;
                i :== (&i + 1)
               OD" 
  ensures_utp "&j =\<^sub>u (\<guillemotleft>a\<guillemotright> + 1)div \<guillemotleft>2\<guillemotright>"
  vcg_sp
   apply presburger+
  done   

program_spec even_count_gen'_sp_H1_H3:
  assumes_utp "\<guillemotleft>a\<guillemotright> >\<^sub>u 0"
  prog_utp    "i :== \<guillemotleft>0::int\<guillemotright>;
               j :== 0 ; 
               INVAR  (&j =\<^sub>u (&i + 1) div 2 \<and> &i \<le>\<^sub>u \<guillemotleft>a\<guillemotright>) 
               VRT (uop nat (\<guillemotleft>a\<guillemotright> - &i))
               WHILE &i <\<^sub>u \<guillemotleft>a\<guillemotright>
               DO
                 IF &i mod 2 =\<^sub>u 0 
                 THEN j :== (&j + 1)
                 ELSE SKIP 
                 FI;
                i :== (&i + 1)
               OD" 
  ensures_utp "&j =\<^sub>u (\<guillemotleft>a\<guillemotright> + 1)div 2"
  vcg_sp
   apply simp
  apply presburger
  done

subsection \<open>sqrt program\<close>
  
definition Isqrt :: "int \<Rightarrow> int \<Rightarrow> bool" 
   where "Isqrt n\<^sub>0 r \<equiv> 0\<le>r \<and> (r-1)\<^sup>2 \<le> n\<^sub>0" 
     
lemma Isqrt_aux:
  "0 \<le> n\<^sub>0 \<Longrightarrow> Isqrt n\<^sub>0 1"
  "\<lbrakk>0 \<le> n\<^sub>0; r * r \<le> n\<^sub>0; Isqrt n\<^sub>0 r\<rbrakk> \<Longrightarrow> Isqrt n\<^sub>0 (r + 1)"
  "\<lbrakk>0 \<le> n\<^sub>0; \<not> r * r \<le> n\<^sub>0; Isqrt n\<^sub>0 r\<rbrakk> \<Longrightarrow> (r - 1)\<^sup>2 \<le> n\<^sub>0 \<and> n\<^sub>0 < r\<^sup>2"
  "Isqrt n\<^sub>0 r \<Longrightarrow> r * r \<le> n\<^sub>0 \<Longrightarrow> r\<le>n\<^sub>0"
  "\<lbrakk>0 \<le> n\<^sub>0; \<not> r * r \<le> n\<^sub>0; Isqrt n\<^sub>0 r\<rbrakk> \<Longrightarrow> 0 < r"
  apply (auto simp: Isqrt_def power2_eq_square algebra_simps)
  by (smt combine_common_factor mult_right_mono semiring_normalization_rules(3))

program_spec sqrt_prog_correct_sp_H1_H3:
  assumes_utp "0 \<le>\<^sub>u \<guillemotleft>a\<guillemotright>"
  ensures_utp "0\<le>\<^sub>u &r \<and> uop power2 (&r) \<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> \<guillemotleft>a\<guillemotright> <\<^sub>u uop power2 (&r + 1)"
  prog_utp    
  "r :== 1 ; 
   INVAR 0\<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> bop Isqrt \<guillemotleft>a\<guillemotright> (&r)
   VRT  uop nat ((\<guillemotleft>a\<guillemotright> + 1) - &r)
   WHILE (&r * &r \<le>\<^sub>u \<guillemotleft>a\<guillemotright>)
   DO 
    r :== (&r + 1)
   OD;
   r :== (&r - 1)" 
  vcg_sp
  by (simp_all add: Isqrt_aux)
  
program_spec sqrt_prog_correct_wp_H1_H3:
  assumes_utp "0 \<le>\<^sub>u \<guillemotleft>a\<guillemotright>"
  ensures_utp "0\<le>\<^sub>u &r \<and> uop power2 (&r) \<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> \<guillemotleft>a\<guillemotright> <\<^sub>u uop power2 (&r + 1)"
  prog_utp    
  "r :== 1 ; 
   INVAR 0\<le>\<^sub>u \<guillemotleft>a\<guillemotright> \<and> bop Isqrt \<guillemotleft>a\<guillemotright> (&r)
   VRT  uop nat ((\<guillemotleft>a\<guillemotright> + 1) - &r)
   WHILE (&r * &r \<le>\<^sub>u \<guillemotleft>a\<guillemotright>)
   DO 
    r :== (&r + 1)
   OD;
   r :== (&r - 1)" 
  vcg_wp
  by (simp_all add: Isqrt_aux)

subsection \<open>gcd program\<close>

program_spec gcd_correct_sp_H1_H3:
  assumes_utp "\<guillemotleft>b\<guillemotright>>\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0"
  ensures_utp "&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>"
  prog_utp    
  "r :== \<guillemotleft>a\<guillemotright>;
   x :== \<guillemotleft>b\<guillemotright>;
   INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT  uop nat (bop max (&r) (&x))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD" 
   vcg_sp
   apply (simp add: gcd_diff1)
   apply (metis gcd.commute gcd_diff1)
  done

program_spec gcd_correct_wp_H1_H3:
  assumes_utp "&r =\<^sub>u \<guillemotleft>a\<guillemotright> \<and> &x =\<^sub>u \<guillemotleft>b\<guillemotright> \<and> \<guillemotleft>b\<guillemotright>>\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0"
  ensures_utp "&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>"
  prog_utp    
  "INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT  uop nat (bop max (&r) (&x))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD" 
   vcg_wp
   apply (simp add: gcd_diff1)
   apply (metis gcd.commute gcd_diff1)+
  done

subsection \<open>max program\<close> 

program_spec max_program_correct:
  assumes_utp  "uop length \<guillemotleft>a\<guillemotright> \<ge>\<^sub>u1 \<and> &i =\<^sub>u 1 \<and> &r =\<^sub>u bop nth \<guillemotleft>a:: int list\<guillemotright> 0"
  ensures_utp "&r =\<^sub>uuop Max (uop set \<guillemotleft>a\<guillemotright>) "  
  prog_utp  
        "INVAR  0 <\<^sub>u &i \<and> &i \<le>\<^sub>u uop length \<guillemotleft>a\<guillemotright> \<and> &r =\<^sub>u uop Max (uop set (bop take (&i) \<guillemotleft>a\<guillemotright>)) 
         VRT  (uop length \<guillemotleft>a\<guillemotright> - (&i))  
         WHILE \<not>(&i =\<^sub>u uop length \<guillemotleft>a\<guillemotright>) 
         DO 
            IF &r <\<^sub>u  bop nth \<guillemotleft>a\<guillemotright> (&i) 
            THEN r :== bop nth \<guillemotleft>a\<guillemotright> (&i)
            ELSE SKIP
            FI;
            i :== (&i + 1)
         OD"  
   vcg_wp   
  subgoal for _ 
    by (cases a; auto)
  subgoal for _ i'
     apply (simp add: take_Suc_conv_app_nth )
    apply (subst (asm) Max_insert) 
      apply auto
    done
  subgoal for _ i' 
    apply (clarsimp simp: take_Suc_conv_app_nth)  
    apply (cases a, auto)
    done
  subgoal for _ i 
    by (clarsimp simp: take_Suc_conv_app_nth)  
  subgoal for _ i   
    by (clarsimp simp: take_Suc_conv_app_nth)  
  done   


subsection \<open>Talk DEMO\<close>

program_spec Talk_gcd_example:
  assumes_utp "\<guillemotleft>b\<guillemotright>>\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0"
  ensures_utp "&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>"
  prog_utp    
  "r :== \<guillemotleft>a\<guillemotright>;
   x :== \<guillemotleft>b\<guillemotright>;
   INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT  uop nat (bop max (&r) (&x))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD" 
vcg_wp      
   apply (simp add: gcd_diff1)
   apply (metis gcd.commute gcd_diff1)+
  done

lemma EXAMPLE_WITHOUT_frontend:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows
 "\<lbrace>\<guillemotleft>b\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0\<rbrace>
   r :== \<guillemotleft>a\<guillemotright>;
   x :== \<guillemotleft>b\<guillemotright>; 
   INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT uop nat (bop max (&r) (&x))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD
 \<lbrace>&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>\<rbrace>\<^sub>P"
  apply (insert assms)      
  apply (wp) 
(*sledgehammer*)
     apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
      apply (simp_all add: gcd_diff1)
   apply (metis gcd.commute gcd_diff1)+
  done

(*MODULARITY DEMO*)

definition \<open>swap i j xs = xs[i := xs!j, j := xs!i]\<close>

abbreviation
  "swaping a j \<equiv> ((a :== (trop swap (\<guillemotleft>j\<guillemotright>) (\<guillemotleft>j\<guillemotright>-1) (&a)))::_ prog)"

lemma Example_swap_preserves_list_elements:
  assumes "vwb_lens a"
  assumes "vwb_lens b"
  assumes "a \<bowtie> b"
  shows 
    "\<lbrace>\<lambda>aa. ((&a) =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and>\<guillemotleft>j\<guillemotright> <\<^sub>u uop length \<guillemotleft>aa\<guillemotright>)\<rbrace>
      \<lambda>aa. swaping a j;(b):== (\<guillemotleft>0\<guillemotright>)
     \<lbrace>\<lambda>aa. (mset\<^sub>u((&a)) =\<^sub>u mset\<^sub>u(\<guillemotleft>aa\<guillemotright>))\<rbrace>\<^sub>G"
  supply swap_def
  apply (insert assms, rule allI,vcg wp)
  unfolding swap_def
  using less_imp_diff_less mset_swap 
  apply blast
  done          

declare Example_swap_preserves_list_elements[THEN spec, vcg_proved_specs]
declare Example_swap_preserves_list_elements[simplified, THEN spec, vcg_proved_specs]

abbreviation
  "count_down j \<equiv> ((j :== ((&j) - \<guillemotleft>1\<guillemotright>))::_ prog)"

lemma Example_count_down_preserves_list_elements:
  assumes "vwb_lens j" 
  shows  
  "\<lbrace>\<lambda>n. (&j) =\<^sub>u \<guillemotleft>n::nat\<guillemotright> \<and> (&j) \<noteq>\<^sub>u \<guillemotleft>0::nat\<guillemotright>\<rbrace> 
     \<lambda>aa.  count_down j
   \<lbrace>\<lambda>n. (&j) \<le>\<^sub>u \<guillemotleft>n\<guillemotright>\<rbrace>\<^sub>G" 
  by (insert assms, rule allI, vcg wp)

declare Example_count_down_preserves_list_elements[ THEN spec, vcg_proved_specs]
declare Example_count_down_preserves_list_elements[simplified,THEN spec, vcg_proved_specs]

lemma swap_count_down_preserves_list:
  assumes "vwb_lens j" 
  assumes "vwb_lens a"
  assumes "vwb_lens b"
  assumes "(a:: 'b list \<Longrightarrow> 'a) \<bowtie> (j:: nat \<Longrightarrow> 'a)"
  shows  
    "\<lbrace>\<lambda>(aa,n). utp_expr.var (GVAR a) =\<^sub>u \<guillemotleft>aa\<guillemotright> \<and> \<guillemotleft>i\<guillemotright> <\<^sub>u uop length \<guillemotleft>aa\<guillemotright> \<and> 
               utp_expr.var (GVAR j) =\<^sub>u \<guillemotleft>n::nat\<guillemotright> \<and> \<guillemotleft>n\<guillemotright> \<noteq>\<^sub>u \<guillemotleft>0::nat\<guillemotright> \<and> 
               utp_expr.var (LVAR b) =\<^sub>u \<guillemotleft>1\<guillemotright>\<rbrace> 
    \<lambda>aa. MODIFY GVAR a DO swaping(GVAR a) i ;(b:snd_lens):== (\<guillemotleft>0\<guillemotright>)OD;
         MODIFY GVAR j DO count_down (GVAR j) OD
   \<lbrace>\<lambda>(aa,n). utp_expr.var (LVAR b) =\<^sub>u \<guillemotleft>1\<guillemotright> \<and> 
             utp_expr.var (GVAR j) \<le>\<^sub>u \<guillemotleft>n\<guillemotright> \<and> 
             mset\<^sub>u(utp_expr.var (GVAR a)) =\<^sub>u mset\<^sub>u(\<guillemotleft>aa\<guillemotright>)\<rbrace>\<^sub>G" 
  apply (insert assms)
  apply (vcg wp) (*Why wp march et wp_trace ne marche pas!!*)
    (*additional post-processing for proving VCs related to the modify assumption
       TODO: automate this*)
      apply (insert assms)

   apply (subst seq_prog_anti_frame_prog_modifies)
     defer
     apply (subst assigns_prog_antiframe_prog_modifies)
      defer
      apply simp
     apply (subst assigns_prog_antiframe_prog_modifies)
      defer
      apply simp
     apply (simp_all add: DONT_TOUCH_def prog_rep_eq)
   apply rel_simp+
  done

(*DEMO VCG step by step and Program annotation WP*)

lemma EXAMPLE_vcg_step_by_step_wp:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows
 "\<lbrace>\<guillemotleft>b\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0\<rbrace>
   r :== \<guillemotleft>a\<guillemotright>;
   x :== \<guillemotleft>b\<guillemotright>; 
   INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT uop nat (bop max (&r) (&x))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD
 \<lbrace>&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>\<rbrace>\<^sub>P"
  apply (insert assms)   
  apply (hoare_wp_vcg_preprocessing)
   apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
      apply (hoare_wp_rule_apply) 
  unfolding DEFERRED_def
(*sledgehammer*)
     apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
      apply (simp_all add: gcd_diff1)
   apply (metis gcd.commute gcd_diff1)+
  done

(*DEMO VCG step by step and Program annotation SP*)
lemma EXAMPLE_vcg_step_by_step_sp:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows
 "\<lbrace>\<guillemotleft>b\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0\<rbrace>
   r :== \<guillemotleft>a\<guillemotright>;
   x :== \<guillemotleft>b\<guillemotright>; 
   INVAR &r >\<^sub>u0 \<and> &x >\<^sub>u 0 \<and> bop gcd (&r) (&x) =\<^sub>u  bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>
   VRT uop nat (bop max (&r) (&x))
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD
 \<lbrace>&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>\<rbrace>\<^sub>P"
  apply (insert assms)   
  apply (hoare_sp_vcg_preprocessing)
   apply (hoare_sp_rule_apply) 
  apply (hoare_sp_rule_apply) 
  apply (hoare_sp_rule_apply) 
  apply (hoare_sp_rule_apply) 
  apply (hoare_sp_rule_apply) 
  apply (hoare_sp_rule_apply) 
  apply (hoare_sp_rule_apply) 
  apply (hoare_sp_rule_apply) 
  apply (hoare_sp_rule_apply) 
  apply (hoare_sp_rule_apply) 
  apply (hoare_sp_rule_apply) 
         apply (hoare_sp_rule_apply)
         apply (hoare_sp_rule_apply)
         apply (hoare_sp_rule_apply)
         apply (hoare_sp_rule_apply)
apply (hoare_sp_rule_apply)
  unfolding DEFERRED_def
(*sledgehammer*)
     apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
   apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
    apply (simp_all add: gcd_diff1)
  defer
    apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))         
   apply (metis gcd.commute gcd_diff1)+
  done

(*DEMO VCG Proof Annotation*)

lemma EXAMPLE_vcg_step_by_step_proof_annotation:
  assumes "r \<bowtie> x"
  assumes "vwb_lens r" "vwb_lens x" 
  shows
 "\<lbrace>\<guillemotleft>b\<guillemotright> >\<^sub>u 0 \<and> \<guillemotleft>a\<guillemotright>>\<^sub>u 0\<rbrace>
   r :== \<guillemotleft>a::int\<guillemotright>;
   x :== \<guillemotleft>b\<guillemotright>; 
   WHILE \<not>(&r =\<^sub>u &x)
   DO
    IF &r >\<^sub>u &x
    THEN r :== (&r - &x)
    ELSE x :== (&x - &r)
    FI
   OD
 \<lbrace>&r =\<^sub>u &x \<and> &r =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>\<rbrace>\<^sub>P"
  apply (insert assms)   
  apply (hoare_wp_vcg_preprocessing)
   apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
    apply (hoare_wp_rule_apply)
   apply (rule while_hoare_prog_wp[where R= measure and st' = "&\<Sigma>" and 
                invar = "utp_expr.var r >\<^sub>u0 \<and> utp_expr.var x >\<^sub>u 0 \<and> 
                         bop gcd (utp_expr.var r) (utp_expr.var x) =\<^sub>u bop gcd \<guillemotleft>a\<guillemotright> \<guillemotleft>b\<guillemotright>" and
                expression  = "uop nat (bop max (utp_expr.var r) (utp_expr.var x))"])
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  apply (hoare_wp_rule_apply) 
  unfolding DEFERRED_def
(*sledgehammer*)
     apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
  apply ((symbolic_execution ; get_remover?; (vcg_elim_determ beautify_thms)?))
      apply (simp_all add: gcd_diff1)
   apply (metis gcd.commute gcd_diff1)+
  done


end