(******************************************************************************
 * Orca: A Functional Correctness Verifier for Imperative Programs
 *       Based on Isabelle/UTP
 *
 * Copyright (c) 2016-2018 Virginia Tech, USA
 *               2016-2018 Technische Universität München, Germany
 *               2016-2018 University of York, UK
 *               2016-2018 Université Paris-Saclay, Univ. Paris-Sud, France
 *
 * This software may be distributed and modified according to the terms of
 * the GNU Lesser General Public License version 3.0 or any later version.
 * Note that NO WARRANTY is provided.
 *
 * See CONTRIBUTORS, LICENSE and CITATION files for details.
 ******************************************************************************)

chapter \<open>Algebraic laws of programming for the theory of total correctness\<close>

theory algebraic_laws_prog
  imports "../impl/utp_prog_des_lift"
begin

(*sledgehammer_params[(*stop_on_first,*)parallel_subgoals, join_subgoals]*)

section \<open>Helpers\<close>

named_theorems algebraic_laws_prog  

lemma vwb_lens_weak[simp]: 
  "vwb_lens x \<Longrightarrow> weak_lens x"
  by simp  

lemma not_ok_refines_any_H1_H2:
  "z is \<^bold>H \<Longrightarrow> z \<sqsubseteq> (\<not> $ok)"
  by rel_auto

lemma Abs_prog_Rep_prog_ndesign:"\<lbrakk>Abs_prog (\<^bold>N P)\<rbrakk>\<^sub>p = (\<^bold>N P)" 
  by (simp add: Abs_prog_inverse H1_H3_idempotent Healthy_def')

lemmas Abs_prog_Rep_prog_Ncarrier= 
  Abs_prog_Rep_prog_ndesign[folded Ncarrier_ndesigns]

lemma fun_apply_rep_prog:
  "\<lbrakk>f X\<rbrakk>\<^sub>p = (Rep_prog \<circ> f \<circ> Abs_prog) \<lbrakk>X\<rbrakk>\<^sub>p"
  by (simp add: Rep_prog_inverse)  

lemma inf_H1_H3_in_lower (*legacy*):
  "(\<And>x. x \<in> A \<Longrightarrow> x is \<^bold>N) \<Longrightarrow> \<^bold>\<Sqinter>\<^bsub>NDES\<^esub>A \<in> Lower (uthy_order NDES) A" 
  unfolding Lower_def
  apply auto                                               
   apply (rule normal_design_theory_continuous.inf_lower )  
    apply (simp_all add: is_Ncarrier_is_ndesigns subsetI)
  done

lemma inf_H1_H3_is_greatest_lower(*legacy*):
  "(\<And>x. x \<in> A \<Longrightarrow> x is \<^bold>N) \<Longrightarrow> greatest (uthy_order NDES) (\<^bold>\<Sqinter>\<^bsub>NDES\<^esub> A) (Lower (uthy_order NDES) A)"
  unfolding greatest_def
  apply (auto simp add: is_Ncarrier_is_ndesigns)        
    apply (metis Lower_closed is_Healthy_subset_member is_Ncarrier_is_ndesigns utp_order_carrier)
  using inf_H1_H3_in_lower apply auto[1]
  apply (rule normal_design_theory_continuous.weak.inf_greatest)
    apply (auto simp add: Lower_def is_Ncarrier_is_ndesigns )
  done  

lemma sup_H1_H3_in_upper(*legacy*):
  "(\<And>x. x \<in> A \<Longrightarrow> x is \<^bold>N ) \<Longrightarrow> \<^bold>\<Squnion>\<^bsub>NDES\<^esub>A \<in> Upper (uthy_order NDES) A" 
  unfolding Upper_def
  apply auto
   apply (rule normal_design_theory_continuous.sup_upper )  
    apply (simp_all add: is_Ncarrier_is_ndesigns subsetI)
  done

lemma sup_H1_H3_is_least_upper(*legacy*):
  "(\<And>x. x \<in> A \<Longrightarrow> x is \<^bold>N) \<Longrightarrow> least (uthy_order NDES) (\<^bold>\<Squnion>\<^bsub>NDES\<^esub> A) (Upper (uthy_order NDES) A)"
  unfolding least_def
  apply (auto simp add: is_Ncarrier_is_ndesigns)        
    apply (metis Upper_closed is_Healthy_subset_member is_Ncarrier_is_ndesigns utp_order_carrier)
  using sup_H1_H3_in_upper apply auto[1]
  apply (rule normal_design_theory_continuous.weak.sup_least)
    apply (auto simp add: Upper_def is_Ncarrier_is_ndesigns )
  done   

section \<open>Laws for Abort and Magic\<close>

lemma abort_prog_refined_by_any:
  "ABORT \<sqsubseteq> PROG"
  by (simp add: Rep_prog_refine abort.rep_eq)

lemma magic_prog_refines_any:
  "SPEC \<sqsubseteq> MAGIC"
  by (metis H1_below_top Healthy_def Rep_prog_H1_H3_closed Rep_prog_refine magic.rep_eq)

section \<open>Laws for Skip\<close>

text \<open>In this section we introduce the algebraic laws of programming related to the SKIP
      statement.\<close>

lemma skip_prog_left_unit [algebraic_laws_prog]:
  "(SKIP ; P) = P"
  by simp

lemma skip_prog_right_unit[algebraic_laws_prog]:
  "(P ; SKIP) = P"
  by simp

section \<open>Laws for Assignment \<close>

text \<open>In this section we introduce the algebraic laws of programming related to the assignment
      statement.\<close>
lemma subst_abort:
  "\<langle>\<sigma>\<rangle>\<^sub>p; ABORT = ABORT"
  apply (simp add: prog_rep_eq)
  apply rel_simp
  done

lemma usubst_assign_p_apply:
  "\<sigma> \<dagger> \<langle>\<rho>\<rangle>\<^sub>p = \<langle>\<rho> o \<sigma>\<rangle>\<^sub>p"
  by (simp add: prog_rep_eq usubst)

lemma assigns_prog_id [algebraic_laws_prog]: "\<langle>id\<rangle>\<^sub>p = SKIP"
  by (simp add: prog_rep_eq)

lemma assigns_prog_test[algebraic_laws_prog]:
  assumes 1:"mwb_lens x"
  shows     "(x :== \<guillemotleft>u\<guillemotright> ; x :== \<guillemotleft>v\<guillemotright>) = (x :== \<guillemotleft>v\<guillemotright>)"
  using 1
  by (simp add: prog_rep_eq assign_design_test)

lemma assigns_prog_left_subst: (*we need to re express the algebraic laws for substitution*)
  "(x :== u ; P) = (psubst [x \<mapsto>\<^sub>s u] P)"
  by (simp add: prog_rep_eq  usubst H1_H3_impl_H2 Rep_prog_H1_H3_closed assigns_d_comp_ext )

lemma assigns_prog_twice[algebraic_laws_prog]:
  assumes "mwb_lens x" and  "x \<sharp> f"
  shows "(x :== e ; x :== f) = (x :== f)"
  using assms
  by (simp add: prog_rep_eq assign_d_twice)

lemma assigns_prog_commute:
  assumes "x \<bowtie> y" "x \<sharp> f" "y \<sharp> e"
  shows "(x :== e ; y :==  f) = (y :== f ; x :==  e)"
  using assms
  by (simp add: prog_rep_eq assigns_d_commute)

lemma assigns_prog_left_cond_cond[algebraic_laws_prog]: 
  fixes x :: "('a \<Longrightarrow> '\<alpha>)"
  shows "(x :== e ; (IF b THEN P ELSE Q FI)) =
         (IF (b\<lbrakk>e/x\<rbrakk>) THEN (x :== e ; P) ELSE (x :== e ; Q) FI)" 
  by (simp add: prog_rep_eq closure assigns_d_left_cond_d_ndesigns)

lemma assigns_prog_uop1[algebraic_laws_prog]:
  assumes 1: "mwb_lens v"
  shows "(v :== e1 ; v :== (uop F (&v))) = (v :== (uop F e1))"
  by (simp add: prog_rep_eq assigns_d_uop1 assms)

lemma assign_prog_bop1[algebraic_laws_prog]:
  assumes 1: "mwb_lens v" and 2:"v \<sharp> e2"
  shows "(v :== e1 ; v :== (bop bp (&v) e2)) = (v :== (bop bp e1 e2))"
  by (simp add: prog_rep_eq assign_d_bop1 assms)

lemma assign_prog_bop2[algebraic_laws_prog]:
  assumes 1: "mwb_lens v" and 2:"v \<sharp> e2"
  shows "(v :== e1 ; v :== (bop bp e2 (&v))) = (v :== (bop bp e2 e1))"
  by (simp add: prog_rep_eq assign_d_bop2 assms)

lemma assign_prog_trop1[algebraic_laws_prog]:
  assumes 1: "mwb_lens v" and 2:"v \<sharp> e2" and 3:"v \<sharp> e3"
  shows "(v :== e1 ; v :== (trop tp (&v) e2 e3)) =
         (v :== (trop tp e1 e2 e3))"
  by (simp add: prog_rep_eq assign_d_trop1 assms)

lemma assign_prog_trop2[algebraic_laws_prog]:
  assumes 1: "mwb_lens v" and 2:"v \<sharp> e2" and 3:"v \<sharp> e3"
  shows "(v :== e1 ; v :== (trop tp e2 (&v) e3)) =
         (v :== (trop tp e2 e1 e3))"
  by (simp add: prog_rep_eq assign_d_trop2 assms)

lemma assign_prog_trop3[algebraic_laws_prog]:
  assumes 1: "mwb_lens v" and 2:"v \<sharp> e2" and 3:"v \<sharp> e3"
  shows "(v :== e1 ; v :== (trop tp e2 e3 (&v))) =
         (v :== (trop tp e2 e3 e1))"
  by (simp add: prog_rep_eq assign_d_trop3 assms)

lemma assign_prog_qtop1[algebraic_laws_prog]:
  assumes 1: "mwb_lens v" and 2:"v \<sharp> e2" and 3:"v \<sharp> e3" and 4:"v \<sharp> e4"
  shows "(v :== e1 ; v :== (qtop tp (&v) e2 e3 e4)) =
         (v :== (qtop tp e1 e2 e3 e4))"
  by (simp add: prog_rep_eq assign_d_qtop1 assms)

lemma assign_prog_qtop2[algebraic_laws_prog]:
  assumes 1: "mwb_lens v" and 2:"v \<sharp> e2" and 3:"v \<sharp> e3" and 4:"v \<sharp> e4"
  shows "(v :== e1 ; v :== (qtop tp e2 (&v) e3 e4)) =
         (v :== (qtop tp e2 e1 e3 e4))"
  by (simp add: prog_rep_eq assign_d_qtop2 assms)

lemma assign_prog_qtop3[algebraic_laws_prog]:
  assumes 1: "mwb_lens v" and 2:"v \<sharp> e2" and 3:"v \<sharp> e3" and 4:"v \<sharp> e4"
  shows "(v :== e1 ; v :== (qtop tp e2 e3 (&v) e4)) =
         (v :== (qtop tp e2 e3 e1 e4))"
  by (simp add: prog_rep_eq assign_d_qtop3 assms)

lemma assign_prog_qtop4[algebraic_laws_prog]:
  assumes 1: "mwb_lens v" and 2:"v \<sharp> e2" and 3:"v \<sharp> e3" and 4:"v \<sharp> e4"
  shows "(v :== e1 ; v :== (qtop tp e2 e3 e4 (&v))) =
         (v :== (qtop tp e2 e3 e4 e1))"
  by (simp add: prog_rep_eq assign_d_qtop4 assms)

text \<open>In the sequel we define assignment laws proposed by Hoare\<close>

lemma assign_prog_vwb_skip_prog[algebraic_laws_prog]:
  assumes 1: "vwb_lens v"
  shows "(v :== &v) = SKIP"
  by (simp add: assms prog_rep_eq assign_d_vwb_skip_d)

lemma assign_prog_simultaneous:
  assumes  1: "vwb_lens var2"
    and      2: "var1 \<bowtie> var2"
  shows "var1, var2 :== expr, &var2 = (var1 :== expr::'a prog)"
  by (simp add: assms assign_d_simultaneous prog_rep_eq)

lemma assign_prog_seq[algebraic_laws_prog]:
  assumes  1: "vwb_lens var2"
  shows"(var1 :== expr) ; (var2 :== &var2) = (var1 :== expr)"
  by (simp add: assms prog_rep_eq assign_d_seq)

lemma assign_prog_cond_prog_uop[algebraic_laws_prog]: (*needs more laws to be automatic*)
  assumes 1: "weak_lens v"
  shows "IF uop F expr THEN (v :== expr ;  C1) ELSE (v :== expr ; C2) FI =
          v :== expr ; IF uop F (&v) THEN  C1 ELSE C2 FI"
  by (simp add: assms prog_rep_eq closure assign_d_cond_d_uop_ndesigns)

lemma assign_prog_cond_bop1[algebraic_laws_prog]: (*needs more laws to be automatic*)
  assumes 1: "weak_lens v" and 2: "v \<sharp> exp2"
  shows "IF bop bp expr exp2 THEN (v :== expr ; C1) ELSE (v :== expr ; C2) FI =
          v :== expr ; IF bop bp (&v) exp2 THEN C1 ELSE C2 FI"
  by (simp add: assms prog_rep_eq closure assign_d_cond_d_bop1_ndesigns)

lemma assign_prog_cond_bop2[algebraic_laws_prog]: (*needs more laws to be automatic*)
  assumes 1: "weak_lens v" and 2: "v \<sharp> exp2"
  shows "IF bop bp exp2 expr THEN (v :== expr ; C1) ELSE (v :== expr ; C2) FI =
          v :== expr ; IF bop bp  exp2 (&v) THEN  C1 ELSE C2 FI"
  by (simp add: assms prog_rep_eq closure assign_d_cond_d_bop2_ndesigns)


lemma assign_prog_cond_prog_trop1[algebraic_laws_prog]: (*needs more laws to be automatic*)
  assumes 1: "weak_lens v" and 2: "v \<sharp> exp2" and 3: "v \<sharp> exp3"
  shows "IF trop bp expr exp2 exp3 THEN (v :== expr ; C1) ELSE (v :== expr ; C2) FI =
         v :== expr ; IF trop bp (&v) exp2 exp3 THEN  C1 ELSE  C2 FI"
  by (simp add: assms prog_rep_eq closure assign_d_cond_d_trop1_ndesigns)


lemma assign_prog_cond_prog_trop2[algebraic_laws_prog]: (*needs more laws to be automatic*)
  assumes 1: "weak_lens v" and 2: "v \<sharp> exp2" and 3: "v \<sharp> exp3"
  shows "IF trop bp exp2 expr exp3 THEN (v :== expr ; C1) ELSE (v :== expr ; C2) FI =
         v :== expr ; IF trop bp exp2 (&v) exp3 THEN C1 ELSE C2 FI"
  by (simp add: assms prog_rep_eq closure assign_d_cond_d_trop2_ndesigns)


lemma assign_prog_cond_prog_trop3[algebraic_laws_prog]: (*needs more laws to be automatic*)
  assumes 1: "weak_lens v" and 2: "v \<sharp> exp2" and 3: "v \<sharp> exp3"
  shows "IF trop bp exp2 exp3 expr THEN (v :== expr ; C1) ELSE (v :== expr ; C2) FI =
         v :== expr ; IF trop bp exp2 exp3 (&v) THEN  C1 ELSE C2 FI"
  by (simp add: assms prog_rep_eq closure assign_d_cond_d_trop3_ndesigns)

lemma assign_prog_cond_prog_qtop1[algebraic_laws_prog]: (*needs more laws to be automatic*)
  assumes 1: "weak_lens v" and 2: "v \<sharp> exp2" and 3: "v \<sharp> exp3" and 4: "v \<sharp> exp4"
  shows "IF qtop bp expr exp2 exp3 exp4 THEN (v :== expr ; Simpl\<^sub>A\<^sub>B\<^sub>R C1) ELSE (v :== expr ; Simpl\<^sub>A\<^sub>B\<^sub>R C2) FI =
         v :== expr ; IF qtop bp (&v) exp2 exp3 exp4 THEN Simpl\<^sub>A\<^sub>B\<^sub>R C1 ELSE Simpl\<^sub>A\<^sub>B\<^sub>R C2 FI"
  by (simp add: assms prog_rep_eq closure assign_d_cond_d_qtop1_ndesigns)

lemma assign_prog_cond_prog_qtop2[algebraic_laws_prog]: (*needs more laws to be automatic*)
  assumes 1: "weak_lens v" and 2: "v \<sharp> exp2" and 3: "v \<sharp> exp3" and 4:"v \<sharp> exp4"
  shows "IF qtop bp exp2 expr  exp3 exp4 THEN (v :== expr ; C1) ELSE (v :== expr ; C2) FI =
         v :== expr ; IF qtop bp exp2 (&v) exp3 exp4 THEN C1 ELSE C2 FI"
  by (simp add: assms prog_rep_eq closure assign_d_cond_d_qtop2_ndesigns)

lemma assign_prog_cond_prog_qtop3[algebraic_laws_prog]: (*needs more laws to be automatic*)
  assumes 1: "weak_lens v" and 2: "v \<sharp> exp2" and 3: "v \<sharp> exp3" and 4: "v \<sharp> exp4"
  shows "IF qtop bp exp2 exp3 expr exp4 THEN (v :== expr ; C1) ELSE (v :== expr ; C2) FI =
         v :== expr ; IF qtop bp exp2 exp3 (&v) exp4 THEN C1 ELSE  C2 FI"
  by (simp add: assms prog_rep_eq closure assign_d_cond_d_qtop3_ndesigns)

lemma assign_prog_cond_prog_qtop4[algebraic_laws_prog]: (*needs more laws to be automatic*)
  assumes 1: "weak_lens v" and 2: "v \<sharp> exp2" and 3: "v \<sharp> exp3" and 4: "v \<sharp> exp4"
  shows "IF qtop bp exp2 exp3 exp4 expr THEN (v :== expr ; C1) ELSE (v :== expr ; C2) FI =
         v :== expr ; IF qtop bp exp2 exp3 exp4 (&v) THEN C1 ELSE C2 FI"
  by (simp add: assms prog_rep_eq closure assign_d_cond_d_qtop4_ndesigns)

lemma assigns_prog_cond_prog[algebraic_laws_prog]:
  "(IF bexp THEN (v :== exp1) ELSE (v :== exp2) FI) =
   (v :== (exp1 \<triangleleft> bexp \<triangleright> exp2))"
  by (simp add: prog_rep_eq closure assign_d_cond_d_If)

section \<open>Laws for Conditional\<close>

lemma If_prog_refines_prog:
  assumes "spec \<sqsubseteq> passert_prog b ; P"
  assumes "spec \<sqsubseteq> passert_prog (\<not>b) ; Q"
  shows  " spec \<sqsubseteq> IF b THEN P ELSE Q FI"
  using assms 
  apply (simp add: prog_rep_eq)
  apply rel_simp
  apply metis
  done

lemma if_prog_pusbst_distr[usubst]:
  "\<sigma> \<dagger>  (IF b THEN P ELSE Q FI) =  IF \<sigma> \<dagger> b THEN \<sigma> \<dagger> P ELSE \<sigma> \<dagger> Q FI"  
  by (simp add: prog_rep_eq usubst)

lemma if_prog_idem[urel_cond]:
  "IF B THEN P ELSE P FI = P"
  by (simp add: prog_rep_eq urel_cond)

lemma if_prog_symm:
  "IF b THEN P ELSE Q FI = IF \<not>b THEN Q ELSE P FI"
  by pauto

lemma if_prog_assoc:
  "IF b THEN P ELSE IF b THEN Q ELSE R FI FI = 
   IF b THEN IF b THEN P ELSE Q FI ELSE R FI"
  by pauto

lemma if_prog_distr'[urel_cond]:
  "IF b THEN IF b' THEN P ELSE R FI ELSE IF b' THEN Q ELSE R FI FI =
   IF b' THEN IF b THEN P ELSE Q FI ELSE R FI"
  by pauto

lemma if_prog_true:
  "IF true THEN P ELSE Q FI = P"
  by pauto

lemma if_prog_false:
  "IF false THEN P ELSE Q FI = Q"
  by pauto

lemma if_prog_L6[urel_cond]:
  "IF b THEN P ELSE IF b THEN Q ELSE R FI FI = 
   IF b THEN P ELSE R FI"
  by pauto

lemma if_prog_conj[urel_cond]:
  "IF b \<and> c THEN P ELSE Q FI = IF b THEN IF c THEN P ELSE Q FI ELSE Q FI"
  by pauto

theorem if_prog_12:
  "IF bexp1 THEN (IF bexp2 THEN C1 ELSE C3 FI) ELSE (IF bexp3 THEN C2 ELSE C3 FI) FI = 
   IF (bexp2 \<triangleleft>bexp1\<triangleright>bexp3) THEN  (IF bexp1 THEN C1 ELSE C2 FI) ELSE C3 FI"
  by pauto

lemma comp_prog_if_prog_left_distr:
  "((IF b THEN P ELSE Q FI); R) = (IF b THEN (P ; R) ELSE (Q ; R) FI)"
  by pauto

lemma comp_prog_if_prog_skip_prog_distr_right: 
    "(IF b THEN SKIP ELSE Q1 FI ; IF b THEN SKIP ELSE Q2 FI) = 
     IF b THEN SKIP ELSE Q1; IF b THEN SKIP ELSE Q2 FI FI"
  by pauto

lemma if_prog_monoI:
  "P\<^sub>1 \<sqsubseteq> P\<^sub>2 \<Longrightarrow> Q\<^sub>1 \<sqsubseteq> Q\<^sub>2 \<Longrightarrow> (IF b THEN P\<^sub>1 ELSE Q\<^sub>1 FI) \<sqsubseteq> (IF b THEN P\<^sub>2 ELSE Q\<^sub>2 FI)"
  by (simp add: prog_rep_eq cond_mono)

lemma If_refines_intro:
  " \<not>b \<longrightarrow> z \<sqsubseteq> Q\<Longrightarrow> b\<longrightarrow> z \<sqsubseteq> P \<Longrightarrow> z \<sqsubseteq> (if b then P else Q)"
  by simp

lemma \<open>vwb_lens x \<Longrightarrow> 
       (x :== \<guillemotleft>c1\<guillemotright> ; IF \<not>(&x =\<^sub>u \<guillemotleft>c1\<guillemotright>) THEN SKIP ELSE x :== \<guillemotleft>c2\<guillemotright> ; Q FI) = 
       (x :== \<guillemotleft>c1\<guillemotright> ; IF false THEN SKIP ELSE x :== \<guillemotleft>c2\<guillemotright> ; Q FI)\<close>
  apply (simp only: if_prog_false)
  apply (smt assigns_prog_left_cond_cond if_prog_symm if_prog_true subst_eq_upred subst_lit ulit_eq usubst_cancel vwb_lens_weak)
  done
lemma  \<open>vwb_lens x \<Longrightarrow>
       (x :== \<guillemotleft>c1\<guillemotright> ; IF \<not>(&x =\<^sub>u \<guillemotleft>c1\<guillemotright>) THEN SKIP ELSE x :== \<guillemotleft>c2\<guillemotright> ; Q FI) = 
       (x :== \<guillemotleft>c2\<guillemotright> ; Q )\<close>
  apply (subst assigns_prog_left_cond_cond)
  apply (subst assigns_prog_left_subst)
  apply (subst assigns_prog_left_subst)
  apply (subst assigns_prog_left_subst)
  apply (subst assigns_prog_left_subst)
  apply ptransfer
  using [[simp_trace]]
  apply pred_simp
  done
lemma blahblahblah: \<open>vwb_lens x \<Longrightarrow> 
       (x :== \<guillemotleft>c1\<guillemotright> ; IF \<not>(&x =\<^sub>u \<guillemotleft>c1\<guillemotright>) THEN SKIP ELSE x :== \<guillemotleft>c2\<guillemotright> ; Q FI) = 
       ( x :== \<guillemotleft>c2\<guillemotright> ; Q )\<close>
    by (smt assigns_prog_left_cond_cond assigns_prog_left_subst assigns_prog_test if_prog_symm if_prog_true psubst_seq subst_eq_upred subst_lit ulit_eq usubst_cancel vwb_lens_mwb vwb_lens_weak)
lemma  blahblahblah1: \<open>vwb_lens x \<Longrightarrow>
       (x :== \<guillemotleft>Some e\<guillemotright> ; IF (&x =\<^sub>u \<guillemotleft>None\<guillemotright>) THEN SKIP ELSE x :== \<guillemotleft>None\<guillemotright> ; Q FI) = 
       (x :== \<guillemotleft>None\<guillemotright> ; Q )\<close>
  
  by (smt assigns_prog_left_cond_cond assigns_prog_left_subst assigns_prog_test if_prog_false option.distinct(1) psubst_seq subst_eq_upred subst_lit ulit_neq usubst_cancel vwb_lens_mwb vwb_lens_weak)
  

section \<open>Laws for Sequential Composition \<close>

lemma seq_prog_monoI:
  "P\<^sub>1 \<sqsubseteq> P\<^sub>2 \<Longrightarrow> Q\<^sub>1 \<sqsubseteq> Q\<^sub>2  \<Longrightarrow> (P\<^sub>1 ; Q\<^sub>1) \<sqsubseteq> (P\<^sub>2 ; Q\<^sub>2)"
  by (simp add: prog_rep_eq seqr_mono)

lemma seq_psubst_prog:
  "\<langle>\<rho>\<rangle>\<^sub>p ; P = (\<rho> \<dagger> P)"   
  by (simp add: prog_rep_eq usubst Rep_prog_H1_H3_closed [THEN H1_H2_impl_H1 [OF H1_H3_impl_H2]])

lemma seq_prog_assoc: "P ; (Q ; R) = (P ; Q) ; R"
  by (simp add: prog_rep_eq seqr_assoc)

section \<open>Monotonic Laws\<close>


lemma if_prog_mono:
  "mono (\<lambda>X. IF b THEN P ; X ELSE Q FI)"
  by (auto intro: monoI seq_prog_monoI if_prog_monoI) 

lemma Mono_progI:
  "(\<And>x y .x \<sqsubseteq> y \<Longrightarrow> f x \<sqsubseteq> f y) \<Longrightarrow> mono_prog f"
  apply ptransfer
  apply (rule Mono_utp_orderI)
  apply (metis Abs_prog_Rep_prog_Ncarrier Healthy_if comp_apply)
  done

lemma Mono_progD:
  "mono_prog f \<Longrightarrow> x \<sqsubseteq> y \<Longrightarrow> f x \<sqsubseteq> f y"
  apply ptransfer
  apply (drule Mono_utp_orderD)
  unfolding fun_apply_rep_prog mono_prog_def
     apply simp_all
   apply (metis Healthy_def' Healthy_if Rep_prog_H1_H3_closed ndes_hcond_def)+
  done

lemma Mono_progE:
  "mono_prog f \<Longrightarrow> x \<sqsubseteq> y \<Longrightarrow> (f x \<sqsubseteq> f y \<Longrightarrow> thesis) \<Longrightarrow> thesis"
  apply ptransfer
  unfolding fun_apply_rep_prog mono_prog_def
  using Mono_utp_orderD Rep_prog_H1_H3_closed is_Ncarrier_is_ndesigns
  by blast

lemma mono_Monotone_prog: 
  assumes M:"mono F"
  shows "mono_prog F"
  by (simp add: Mono_progI assms monoD)

lemma if_prog_mono_prog:
  "mono_prog (\<lambda>X. IF b THEN P ; X ELSE Q FI)"
  by (simp add: if_prog_mono mono_Monotone_prog)

section \<open>Laws for Lattices\<close> 

subsection\<open>alternative definitions for least and upper\<close>

lemma least_prog_alt_def:
  "least_prog l A = (l \<in> A \<and> (ALL x : A. l \<sqsubseteq> x))"
  unfolding  prog_rep_eq least_def image_def Ball_def
  by (auto simp add: Rep_prog_inject Rep_prog_H1_H3_closed is_Ncarrier_is_ndesigns)

lemma Upper_prog_alt_def: 
  "Upper_prog A = {u. (ALL x. x \<in> A \<longrightarrow> x \<sqsubseteq> u)}"  
  unfolding   prog_rep_eq Upper_def image_def Ball_def Upper_prog_def
  apply auto
   apply (metis (mono_tags, lifting) Abs_prog_Rep_prog_Ncarrier Healthy_if 
      Rep_prog_H1_H3_closed is_Ncarrier_is_ndesigns)
  apply (smt IntE IntI Rep_prog Rep_prog_inverse is_Ncarrier_is_ndesigns mem_Collect_eq)
  done  

lemma least_prog_unique:
  "least_prog y A\<Longrightarrow> least_prog x A \<Longrightarrow> x = y"
  by (simp add: Rep_prog_eq least_prog.rep_eq ndes_utp_theory.least_unique)

lemma Lower_prog_alt_def:
  "Lower_prog A = {l. (ALL x. x \<in> A \<longrightarrow> l \<sqsubseteq> x)}"  
  unfolding   prog_rep_eq Lower_def image_def Ball_def Lower_prog_def
  apply auto
   apply (metis (mono_tags, lifting) Abs_prog_Rep_prog_Ncarrier Healthy_if Rep_prog_H1_H3_closed is_Ncarrier_is_ndesigns)
  apply (smt IntE IntI Rep_prog Rep_prog_inverse is_Ncarrier_is_ndesigns mem_Collect_eq)
  done

subsection \<open>alternative definitions for greatest and lower\<close>  

lemma greatest_prog_alt_def: 
  "greatest_prog g A =  (g \<in> A \<and> (ALL x.  x \<in> A \<longrightarrow> x \<sqsubseteq> g))"
  unfolding  prog_rep_eq greatest_def image_def Ball_def
  by (auto simp add: Rep_prog_inject Rep_prog_H1_H3_closed is_Ncarrier_is_ndesigns)

lemma greatest_prog_unique:
  "greatest_prog y A\<Longrightarrow> greatest_prog x A \<Longrightarrow> x = y"
  by (simp add: Rep_prog_eq greatest_prog.rep_eq ndes_utp_theory.greatest_unique)

subsection \<open>Laws for Sup\<close>  

lemma sup_prog_empty:
  "\<Squnion>\<^sub>p{} = ABORT"  
  by (simp add: prog_rep_eq utp_theory_ndes_bot_is_true)

lemma sup_prog_univ:
  "\<Squnion>\<^sub>pUNIV = MAGIC"    
  apply (simp add: prog_rep_eq sup_def least_def Upper_def)    
  apply (rule someI2_ex)
   apply (rule exI[where x = "(\<not> $ok)"])
   apply auto[]
     apply (metis H1_below_top Healthy_if ndes_hcond_def)
    apply rel_simp
   apply (metis des_top_is_H1_H3 is_Ncarrier_is_ndesigns magic.rep_eq range_eqI)
  apply auto[]
  apply (metis (no_types, hide_lams) H1_below_top Healthy_if antisym_conv des_top_ndes_def 
      magic.rep_eq ndes_hcond_def ndesign_is_healthy_NDES range_eqI)     
  done

lemma sup_prog_least:
  "(\<And>x. x \<in> A \<Longrightarrow> x \<sqsubseteq> z) \<Longrightarrow> \<Squnion>\<^sub>p A \<sqsubseteq> z" 
  apply (simp only: prog_rep_eq)
  apply (rule normal_design_theory_continuous.weak.sup_least)
  using  Rep_prog
    apply auto  
  done

lemma sup_prog_upper:
  "x \<in> A \<Longrightarrow> x \<sqsubseteq> \<Squnion>\<^sub>p A " 
  apply (simp only: prog_rep_eq)
  apply (metis (mono_tags, lifting) Rep_prog ball_imageD image_subsetI normal_design_theory_continuous.sup_upper)
  done

lemma sup_prog_in_upper_prog:
  "(\<Squnion>\<^sub>p A) \<in> (Upper_prog A)"
  unfolding Upper_prog_alt_def
  by (auto intro!: sup_prog_upper)

lemma sup_prog_is_least_upper_prog:
  "least_prog (\<Squnion>\<^sub>p A) (Upper_prog A)"
  unfolding least_prog_alt_def Upper_prog_alt_def
proof (auto simp: sup_prog_in_upper_prog, goal_cases)
  case (1 x)
  then show ?case by (rule sup_prog_upper) 
next
  case (2 x)
  then show ?case by (auto intro!: sup_prog_least)
qed

lemma sup_prog_alt_def:
  "(\<Squnion>\<^sub>p A) = (SOME x. least_prog x (Upper_prog A))" 
  apply (rule someI2_ex)
  using sup_prog_is_least_upper_prog
   apply blast
  apply (simp add: least_prog_unique sup_prog_is_least_upper_prog)
  done

subsection \<open>Laws for Inf\<close>  

lemma inf_prog_empty:
  "\<Sqinter>\<^sub>p{} = MAGIC"  
  by (simp add: prog_rep_eq utp_theory_ndes_top_is_not_ok)  

lemma inf_prog_univ:
  "\<Sqinter>\<^sub>pUNIV = ABORT"    
proof (simp add: prog_rep_eq, subst utp_theory_continuous.healthy_inf_def, goal_cases)
  case 1
  then show ?case 
    by (simp add: normal_design_theory_continuous.utp_theory_continuous_axioms) 
next
  case 2
  then show ?case 
    by (simp add: Rep_prog_H1_H3_closed image_subsetI is_Ncarrier_is_ndesigns) 
next
  case 3
  then show ?case 
    by (metis Sup_upper abort.rep_eq all_not_in_conv rangeI utp_pred_laws.top.extremum_uniqueI) 
qed

lemma inf_prog_greatest:
  assumes "\<And>x. x \<in> A \<Longrightarrow> z \<sqsubseteq> x"
  shows "z \<sqsubseteq> \<Sqinter>\<^sub>p A"
proof (simp add: prog_rep_eq, subst utp_theory_continuous.healthy_inf_def, goal_cases)
  case 1                                                     
  then show ?case 
    by (simp add: normal_design_theory_continuous.utp_theory_continuous_axioms)
next
  case 2
  then show ?case 
    by (simp add: Rep_prog_H1_H3_closed image_subsetI is_Ncarrier_is_ndesigns)
next
  case 3 
  then show ?case  
  proof (rule If_refines_intro, goal_cases)
    case 1
    then show ?case apply (subst image_is_empty)
      apply (rule impI)
      apply (rule cSUP_least)
       apply (assumption)    
      apply (rule assms[unfolded less_eq_prog.rep_eq])
      apply assumption
      done
  next
    case 2
    then show ?case 
      apply (subst utp_theory_ndes_top_is_not_ok)
      apply (rule impI)
      apply (rule  H1_H3_impl_H2[THEN not_ok_refines_any_H1_H2])
      apply (rule Rep_prog_H1_H3_closed)
      done
  qed
qed

lemma inf_prog_lower:
  "x \<in> A \<Longrightarrow> \<Sqinter>\<^sub>pA \<sqsubseteq> x"
proof (simp add: prog_rep_eq,subst utp_theory_continuous.healthy_inf_def, goal_cases)
  case 1
  then show ?case   
    by (simp add: normal_design_theory_continuous.utp_theory_continuous_axioms)
next
  case 2
  then show ?case
    by (simp add: Rep_prog_H1_H3_closed image_subsetI is_Ncarrier_is_ndesigns) 
next
  case 3
  assume 1:"x\<in> A"
  then show ?case using SUP_upper 
    by fastforce
qed

lemma inf_prog_lower2:
  "u \<in> A \<Longrightarrow> u \<sqsubseteq> v \<Longrightarrow> (\<Sqinter>\<^sub>p A) \<sqsubseteq> v"
  using inf_prog_lower [of u A] by auto

lemma inf_prog_in_lower_prog:
  "(\<Sqinter>\<^sub>p A) \<in> (Lower_prog A)"
  unfolding Lower_prog_alt_def
  by (auto intro!: inf_prog_lower)

lemma inf_prog_is_greatest_lower_prog:
  "greatest_prog (\<Sqinter>\<^sub>p A) (Lower_prog A)"
  unfolding greatest_prog_alt_def Lower_prog_alt_def  
proof (auto, goal_cases)
  case (1 x)
  then show ?case by (rule inf_prog_lower) 
next
  case (2 x)
  then show ?case by (auto intro!: inf_prog_greatest) 
qed

lemma inf_eqI:
  "\<forall>xa. xa \<in> A \<longrightarrow> x \<sqsubseteq> xa \<Longrightarrow> \<forall>xa. (\<forall>x. x \<in> A \<longrightarrow> xa \<sqsubseteq> x) \<longrightarrow> xa \<sqsubseteq> x \<Longrightarrow> \<Sqinter>\<^sub>pA = x"
  by (simp add: antisym inf_prog_greatest inf_prog_lower)

lemma inf_prog_alt_def:
  "(\<Sqinter>\<^sub>p A) = (SOME x. greatest_prog x (Lower_prog A))" 
proof (rule someI2_ex, goal_cases)
  case 1
  then show ?case using inf_prog_is_greatest_lower_prog by auto
next
  case (2 x)
  then show ?case  unfolding greatest_prog_alt_def Lower_prog_alt_def by (simp add: inf_eqI)
qed

subsection \<open>Laws for join\<close> 

lemma utp_join_prog_alt_def:
  "a \<squnion>\<^sub>p b = \<Squnion>\<^sub>p{a, b}"
  by (simp add: prog_rep_eq join_def)

lemma le_join_progI1: 
  "s \<sqsubseteq> x \<Longrightarrow> s \<sqsubseteq> x \<squnion>\<^sub>p y"
  apply (rule order_trans)
   apply auto
  unfolding utp_join_prog_alt_def
  apply (rule sup_prog_upper)
  apply auto
  done

lemma le_join_progI2: 
  "s \<sqsubseteq> y \<Longrightarrow> s \<sqsubseteq> x \<squnion>\<^sub>p y"
  apply (rule order_trans)
   apply auto
  unfolding utp_join_prog_alt_def
  apply (rule sup_prog_upper)
  apply auto
  done

lemma le_join_progI: 
  "s \<sqsubseteq> x \<Longrightarrow> s \<sqsubseteq> y \<Longrightarrow>  s \<sqsubseteq> x \<squnion>\<^sub>p y"
  apply (rule order_trans)
   apply auto
  unfolding utp_join_prog_alt_def
  apply (rule sup_prog_upper)
  apply auto
  done

subsection \<open>Laws for meet\<close>   

lemma meet_prog_alt_def:
  "a \<sqinter>\<^sub>p b = \<Sqinter>\<^sub>p{a, b}"
  by (simp add: prog_rep_eq meet_def)

lemma le_meet_prog_abrI1:
  "a \<sqsubseteq> x \<Longrightarrow> a \<sqinter>\<^sub>p b \<sqsubseteq> x"
  apply (rule order_trans)
   apply auto
  unfolding meet_prog_alt_def
  apply (rule inf_prog_lower)
  apply auto
  done

lemma le_meet_prog_abrI2:
  "b \<sqsubseteq> x \<Longrightarrow> a \<sqinter>\<^sub>p b \<sqsubseteq> x"
  apply (rule order_trans)
   apply auto
  unfolding meet_prog_alt_def
  apply (rule inf_prog_lower)
  apply auto
  done

lemma le_meet_prog_abrI: 
  "a \<sqsubseteq> x \<Longrightarrow> b \<sqsubseteq> x \<Longrightarrow> a \<sqinter>\<^sub>p b \<sqsubseteq> x"
  apply (rule order_trans)
   apply auto
  unfolding meet_prog_alt_def
  apply (rule inf_prog_lower)
  apply auto
  done 

section \<open>laws for Fixed points\<close>   

subsection \<open>greatest fixed points laws\<close> 

lemma gfp_prog_alt_def:
  "\<nu>\<^sub>p F = \<Squnion>\<^sub>p{x. x \<sqsubseteq> F x}"
  apply (simp only: prog_rep_eq)
  unfolding GFP_def sup_def image_def
  apply auto
  apply (rule someI2_ex)
   apply (rule exI[where x = "\<^bold>\<Squnion>\<^bsub>NDES\<^esub> {y. \<exists>x. \<lbrakk>x\<rbrakk>\<^sub>p \<sqsubseteq> \<lbrakk>F x\<rbrakk>\<^sub>p \<and> y = \<lbrakk>x\<rbrakk>\<^sub>p}"])
    (*Test Timeout bug of sledgehammer para*)
   apply (rule sup_H1_H3_is_least_upper)
  using Rep_prog 
   apply fastforce
  apply (rule someI2_ex)
   apply (rule exI[where x = "\<^bold>\<Squnion>\<^bsub>NDES\<^esub> {u. (u is \<H>\<^bsub>NDES\<^esub>) \<and> u \<sqsubseteq> \<lbrakk>F (Abs_prog u)\<rbrakk>\<^sub>p}"])
   apply (rule sup_H1_H3_is_least_upper)
   apply (simp add: is_Ncarrier_is_ndesigns)
  apply (rule ndes_utp_theory.least_unique)
   apply auto
  unfolding least_def Upper_def
  apply auto
   apply (simp add: Rep_prog_inverse)
  unfolding Ball_def
  apply auto
  apply (metis Abs_prog_Rep_prog_Ncarrier Healthy_if)
  done

lemma gfp_prog_unfold:
  assumes M : "mono_prog F"
  shows "\<nu>\<^sub>p F = F (\<nu>\<^sub>p F)"
  apply (simp only: prog_rep_eq) 
  unfolding fun_apply_rep_prog prec_gfp_prog.rep_eq
  apply (subst (1) normal_design_theory_continuous.GFP_healthy_comp) 
  apply (subst (2) normal_design_theory_continuous.GFP_healthy_comp)  
  using M
  apply (metis (mono_tags, lifting) Pi_I Rep_prog mono_prog.rep_eq normal_design_theory_continuous.GFP_healthy_comp normal_design_theory_continuous.GFP_unfold o_def)
  done

lemma gfp_prog_const:
  "\<nu>\<^sub>p (\<lambda>x. t) = t"
  by (simp add: Mono_progI gfp_prog_unfold)

lemma gfp_prog_idem:
  "mono_prog F \<Longrightarrow> idem_prog F \<Longrightarrow> \<nu>\<^sub>p F = F MAGIC"
proof (ptransfer, 
    subst normal_design_theory_continuous.weak.GFP_idem[unfolded utp_theory_ndes_top_is_not_ok],
    goal_cases)
  case 1
  then show ?case using Rep_prog by fastforce
next
  case 2
  then show ?case by simp
next
  case 3
  then show ?case by simp
next
  case 4
  then show ?case  unfolding fun_apply_rep_prog by (simp add: magic.rep_eq)
qed

lemma gfp_prog_least:
  "(\<And>z.  z \<sqsubseteq> F z \<Longrightarrow> z \<sqsubseteq> x ) \<Longrightarrow> \<nu>\<^sub>p F \<sqsubseteq> x"
  by (metis (no_types, lifting) CollectD gfp_prog_alt_def sup_prog_least)

lemma gfp_prog_upperbound:
  "x \<sqsubseteq> F x \<Longrightarrow> x \<sqsubseteq> \<nu>\<^sub>p F"
  by (simp add: gfp_prog_alt_def sup_prog_upper)

lemma gfp_prog_mono:
  "(\<And>x. F x \<sqsubseteq> G x) \<Longrightarrow> \<nu>\<^sub>p F \<sqsubseteq> \<nu>\<^sub>p G"
  apply (rule gfp_prog_upperbound[THEN gfp_prog_least])
  apply (blast intro: order_trans)  
  done

lemma gfp_prog_upperboundI:
  "x = F x \<Longrightarrow> x \<sqsubseteq> \<nu>\<^sub>p F"
  by (simp add: gfp_prog_upperbound)

lemma gfp_prog_fixed_point:
  "mono_prog F \<Longrightarrow> F (\<nu>\<^sub>p F) = \<nu>\<^sub>p F"
  apply (subst (2) gfp_prog_unfold)
   apply auto
  done

lemma gfp_prog_ordinal_induct [case_names mono step union]:
  assumes M: "mono_prog F"
  assumes P_f:   "\<And>S. P S \<Longrightarrow> \<nu>\<^sub>p F \<sqsubseteq> S \<Longrightarrow> P (F S)"
  assumes P_Union: "\<And>M. \<forall>S\<in>M. P S \<Longrightarrow> P (inf_prog M)"
  shows "P (\<nu>\<^sub>p F)"
proof -
  let ?M = "{S. \<nu>\<^sub>p F \<sqsubseteq> S \<and> P S}"
  from P_Union have "P (inf_prog ?M)" by simp
  also have "inf_prog ?M = \<nu>\<^sub>p F"
  proof (rule antisym)
    show "\<nu>\<^sub>p F \<sqsubseteq> inf_prog ?M "
      by (blast intro: inf_prog_greatest)
    then have "F (\<nu>\<^sub>p F) \<sqsubseteq> F(inf_prog ?M)"
      using M Mono_progE by auto
    then have "\<nu>\<^sub>p F \<sqsubseteq> F (inf_prog ?M)"
      by (simp add: M gfp_prog_fixed_point)
    then have "F (inf_prog ?M) \<in> ?M"
      using P_Union by simp (intro P_f inf_prog_greatest, auto)
    then have "inf_prog ?M \<sqsubseteq> F (inf_prog ?M) "
      by (rule inf_prog_lower)
    then show "inf_prog ?M \<sqsubseteq> \<nu>\<^sub>p F"
      by (rule gfp_prog_upperbound)
  qed
  finally show ?thesis .
qed

subsection \<open>least fixed points laws\<close> 

lemma lfp_prog_alt_def:
  "\<mu>\<^sub>p F = \<Sqinter>\<^sub>p{x.  F x \<sqsubseteq> x}"      
  apply (simp only: prog_rep_eq)
  unfolding LFP_def inf_def image_def
  apply auto
  apply (rule someI2_ex)
   apply (rule exI[where x = "\<^bold>\<Sqinter>\<^bsub>NDES\<^esub>{y. \<exists>x. \<lbrakk>F x\<rbrakk>\<^sub>p \<sqsubseteq> \<lbrakk>x\<rbrakk>\<^sub>p \<and> y = \<lbrakk>x\<rbrakk>\<^sub>p}"])
   apply (rule inf_H1_H3_is_greatest_lower)
  using Rep_prog 
   apply fastforce
  apply (rule someI2_ex)
   apply (rule exI[where x="\<^bold>\<Sqinter>\<^bsub>NDES\<^esub>{u. (u is \<H>\<^bsub>NDES\<^esub>) \<and> \<lbrakk>F (Abs_prog u)\<rbrakk>\<^sub>p \<sqsubseteq> u}"])
   apply (rule inf_H1_H3_is_greatest_lower)
   apply (simp add: is_Ncarrier_is_ndesigns)
  apply (rule  ndes_utp_theory.greatest_unique)
   apply auto
  unfolding greatest_def Lower_def Ball_def
  apply auto
   apply (simp add: Rep_prog_inverse)
  apply (metis Abs_prog_Rep_prog_Ncarrier Healthy_if)
  done

lemma lfp_prog_unfold: 
  assumes M:"mono_prog F"  
  shows "\<mu>\<^sub>p F = F (\<mu>\<^sub>p F)"
proof -
  show ?thesis  
    apply (simp only: prog_rep_eq)
    unfolding fun_apply_rep_prog  prec_lfp_prog.rep_eq 
    apply (subst (1) normal_design_theory_continuous.LFP_healthy_comp) 
    apply (subst (2) normal_design_theory_continuous.LFP_healthy_comp)    
    using M 
    by (metis (mono_tags, lifting) Pi_I Rep_prog mono_prog.rep_eq normal_design_theory_continuous.LFP_healthy_comp normal_design_theory_continuous.LFP_unfold o_def)
qed

lemma lfp_prog_const: "\<mu>\<^sub>p (\<lambda>x. t) = t"
  by (simp add: mono_prog.rep_eq lfp_prog_unfold utp_theory.Mono_utp_orderI)

lemma lfp_prog_idem:
  "mono_prog F \<Longrightarrow> idem_prog F \<Longrightarrow> \<mu>\<^sub>p F = F ABORT"
proof (ptransfer,
    subst normal_design_theory_continuous.weak.LFP_idem[unfolded utp_theory_ndes_bot_is_true],
    goal_cases)
  case 1
  then show ?case using Rep_prog by fastforce
next
  case 2
  then show ?case by simp 
next
  case 3
  then show ?case by simp 
next
  case 4
  then show ?case unfolding fun_apply_rep_prog by (simp add: abort.rep_eq)
qed

lemma lfp_prog_lowerbound:
  "F x \<sqsubseteq> x \<Longrightarrow> \<mu>\<^sub>p F \<sqsubseteq> x"
  by (simp add: inf_prog_lower lfp_prog_alt_def)

lemma lfp_prog_greatest:
  "(\<And>z. F z \<sqsubseteq> z \<Longrightarrow> x \<sqsubseteq> z) \<Longrightarrow> x \<sqsubseteq> \<mu>\<^sub>p F"   
  by (metis (full_types) CollectD inf_prog_greatest lfp_prog_alt_def)

lemma lfp_prog_mono:
  "(\<And>x. F x \<sqsubseteq> G x)\<Longrightarrow> \<mu>\<^sub>p F \<sqsubseteq> \<mu>\<^sub>p G"
  apply (rule lfp_prog_lowerbound[THEN lfp_prog_greatest])
  apply (blast intro: order_trans)  
  done

lemma lfp_prog_fixed_point:
  "mono_prog F \<Longrightarrow>  F (\<mu>\<^sub>p F) = \<mu>\<^sub>p F"
  apply (subst (2) lfp_prog_unfold)
   apply auto
  done

lemma lfp_prog_lowerboundI:
  "F x = x \<Longrightarrow> \<mu>\<^sub>p F \<sqsubseteq> x"
  by (simp add: lfp_prog_lowerbound)

lemma lfp_prog_ordinal_induct [case_names mono step union]:
  assumes M: "mono_prog F"
  assumes P_f:   "\<And>S. P S \<Longrightarrow> S \<sqsubseteq> \<mu>\<^sub>p F \<Longrightarrow> P (F S)"
  assumes P_Union: "\<And>M. \<forall>S\<in>M. P S \<Longrightarrow> P (sup_prog M)"
  shows "P (\<mu>\<^sub>p F)"
proof -
  let ?M = "{S. S \<sqsubseteq> \<mu>\<^sub>p F \<and> P S}"
  from P_Union have "P (sup_prog ?M)" by simp
  also have "sup_prog ?M = \<mu>\<^sub>p F"
  proof (rule antisym)
    show "\<mu>\<^sub>p F \<sqsubseteq> sup_prog ?M "
      apply (rule lfp_prog_lowerbound)
      apply (rule sup_prog_upper)
      apply simp
      apply (subst lfp_prog_unfold[OF M])
      apply auto
       apply (insert M)
       apply (erule Mono_progD)
       apply (rule sup_prog_least)
       apply auto[]
      apply (rule P_f)
       apply (simp add: calculation)
      apply (rule sup_prog_least)
      apply auto
      done
    then have "F (\<mu>\<^sub>p F) \<sqsubseteq> F(sup_prog ?M)"
      using M Mono_progE by auto
    then have "\<mu>\<^sub>p F \<sqsubseteq> F (sup_prog ?M)"
      by (simp add: M lfp_prog_fixed_point)
    then have "F (sup_prog ?M) \<in> ?M"
      apply auto
       apply (rule antisym)
        apply (subst lfp_prog_unfold[OF M])
      using M apply (elim Mono_progD)
        apply (rule sup_prog_least)
        apply auto[]
       apply assumption
      apply (rule P_f)
       apply (simp add: calculation)
      apply (rule sup_prog_least)
      apply auto[]
      done
    then have "sup_prog ?M \<sqsubseteq> F (sup_prog ?M) "
      using \<open>\<mu>\<^sub>p F \<sqsubseteq> F (\<Squnion>\<^sub>p{S. S \<sqsubseteq> \<mu>\<^sub>p F \<and> P S})\<close>
      by (auto intro!: sup_prog_least)
    then show "sup_prog ?M \<sqsubseteq> \<mu>\<^sub>p F"
      using \<open>F (\<Squnion>\<^sub>p{S. S \<sqsubseteq> \<mu>\<^sub>p F \<and> P S}) \<in> {S. S \<sqsubseteq> \<mu>\<^sub>p F \<and> P S}\<close> by auto
  qed
  finally show ?thesis .
qed

lemma lfp_prog_termination_intro:
  assumes "termination":"\<mu>\<^sub>p F = \<nu>\<^sub>p F"
  assumes step:"P \<sqsubseteq> F P"   
  shows "P \<sqsubseteq> \<mu>\<^sub>p F" 
  by (simp add: assms gfp_prog_upperbound)

section \<open>laws for Iterations\<close>

text \<open>In this section we introduce the algebraic laws of programming related to the while
      statement.\<close>

lemma while_lfp_prog_mu_prog:
  "WHILE b DO body OD = (\<mu>\<^sub>p X \<bullet> IF b THEN body ; X ELSE SKIP FI)"
  unfolding  while_lfp_prog_def from_until_lfp_prog_def skip_prog_left_unit
  by simp 

theorem while_lfp_prog_unfold:
  "WHILE b DO body OD = (IF b THEN (body ; WHILE b DO body OD) ELSE SKIP FI)"
proof -  
  have "(WHILE b DO body OD) = (\<mu>\<^sub>p X \<bullet> IF b THEN (body ; X) ELSE SKIP FI)"
    by (simp add: while_lfp_prog_mu_prog)
  also have "... = (IF b THEN (body ; (\<mu>\<^sub>p X \<bullet> IF b THEN (body ; X) ELSE SKIP FI)) ELSE SKIP FI)"
  proof (rule lfp_prog_unfold[OF Mono_progI ], goal_cases)
    case (1 x y)
    then show ?case by (simp add: if_prog_monoI seq_prog_monoI)
  qed  
  also have "... = (IF b THEN (body ; WHILE b DO body OD) ELSE SKIP FI)"
    by (simp add: while_lfp_prog_mu_prog)
  finally show ?thesis .
qed

theorem while_lfp_prog_true: 
  "WHILE true DO body OD = (\<mu>\<^sub>p X \<bullet>  body ; X)" 
  unfolding while_lfp_prog_mu_prog  if_prog_true ..

lemma while_lfp_prog_false:
  "(WHILE false DO body OD) = SKIP"
  unfolding while_lfp_prog_mu_prog if_prog_false
  by (simp add: lfp_prog_const) 

lemma while_lfp_prog_non_termination:
  "WHILE true DO SKIP OD = ABORT"
  unfolding while_lfp_prog_mu_prog if_prog_true 
  by (auto simp add: lfp_prog_idem Mono_progI Rep_prog_inverse idem_prog.rep_eq idempotent_def)

subsection \<open>Iteration laws\<close>

lemma from_until_lfp_prog_def_alt:
  "FROM init UNTIL b DO body OD = init ; WHILE \<not>b DO body OD"
  unfolding from_until_lfp_prog_def while_lfp_prog_mu_prog ..

lemma from_until_lfp_invr_vrt_prog_def_alt:
  "FROM init INVAR invar VRT vari UNTIL b DO body OD = 
   init ; INVAR invar VRT vari WHILE \<not>b DO body OD"
  unfolding from_until_lfp_invr_vrt_prog_def from_until_lfp_prog_def_alt 
            while_lfp_invr_vrt_prog_def
  by simp

lemma from_until_lfp_prog_unfold:
  "FROM init UNTIL b DO body OD = init ; IF \<not>b THEN body ; WHILE \<not>b DO body OD ELSE SKIP FI"
  unfolding from_until_lfp_prog_def_alt 
  using while_lfp_prog_unfold[of "\<not>b" body]
  by simp

lemma from_until_lfp_prog_true: 
  "FROM init UNTIL true DO body OD = init"
  unfolding from_until_lfp_prog_def_alt
  by (simp add: while_lfp_prog_false)

lemma from_until_lfp_prog_non_termination: (*should equal abort situation*)
  "FROM init UNTIL false DO SKIP OD = init ; ABORT"
  unfolding from_until_lfp_prog_def_alt
  by (simp add: while_lfp_prog_non_termination)

lemma do_while_lfp_prog_def_alt:
  "DO body WHILE b OD = body ; WHILE b DO body OD"
  unfolding do_while_lfp_prog_def from_until_lfp_prog_def_alt
  by simp

lemma do_while_invar_vrt_lfp_prog_def_alt:
  "DO body INVAR invar VRT vari WHILE b OD = 
   body ; INVAR invar VRT vari WHILE b DO body OD"
  unfolding do_while_lfp_invr_vrt_prog_def do_while_lfp_prog_def_alt while_lfp_invr_vrt_prog_def
  by simp 

lemma do_while_lfp_prog_unfold: 
  "DO body WHILE b OD = body ; IF b THEN body; WHILE b DO body OD ELSE SKIP FI"
  unfolding do_while_lfp_prog_def_alt
  using while_lfp_prog_unfold[of b body]
  by simp

lemma do_while_lfp_prog_false:
  "DO body WHILE false OD = body"
  unfolding do_while_lfp_prog_def_alt
  by (simp add: while_lfp_prog_false)

lemma do_while_lfp_prog_non_termination:
  "DO SKIP WHILE true OD = ABORT" 
  unfolding do_while_lfp_prog_def_alt
  by (simp add: while_lfp_prog_non_termination)

lemma for_lfp_prog_def_alt:
  "FOR (init, b, incr) DO body OD = 
   init ; WHILE b DO body ; incr OD"
  unfolding for_lfp_prog_def from_until_lfp_prog_def_alt
  by simp

lemma for_lfp_invar_vrt_prog_def_alt:
  "FOR (init, b, incr) INVAR invar VRT vari DO body OD = 
   init ; INVAR invar VRT vari WHILE b DO body ; incr OD"
  unfolding for_lfp_invr_vrt_prog_def for_lfp_prog_def_alt while_lfp_invr_vrt_prog_def
  by simp

lemma for_lfp_prog_unfold: 
  "FOR (init, b, incr) DO body OD = 
   init; IF b THEN body; incr; WHILE b DO body; incr OD ELSE SKIP FI"
  unfolding for_lfp_prog_def_alt 
  using while_lfp_prog_unfold [of b "body;incr"] seq_prog_assoc[THEN HOL.sym]
  by metis

lemma for_lfp_prog_false:
  "FOR (init, false, incr) DO body OD = init"
  unfolding for_lfp_prog_def_alt
  by (simp add: while_lfp_prog_false)

lemma for_lfp_prog_non_termination:
  "FOR (init, true, incr) DO SKIP OD = init; WHILE true DO incr OD"
  unfolding for_lfp_prog_def_alt
  by (simp add: while_lfp_prog_non_termination)

section \<open>laws for assume and assert\<close>

lemma assume_twice[algebraic_laws_prog]: "((b\<^sup>\<top>\<^sup>P) ; (c\<^sup>\<top>\<^sup>P)) = (b \<and> c)\<^sup>\<top>\<^sup>P"
  by (simp add: prog_rep_eq) (metis assume_d_twice assume_des_def)  

lemma assert_twice[algebraic_laws_prog]: "(b\<^sub>\<bottom>\<^sub>P ; (c\<^sub>\<bottom>\<^sub>P)) = (b \<and> c)\<^sub>\<bottom>\<^sub>P"
  by (simp add: prog_rep_eq) (metis assert_d_twice assert_des_def) 

section \<open>laws for frame and antiframe \<close>

subsection \<open>laws for antiframe \<close>

lemma antiframe_test:
  "vwb_lens a \<Longrightarrow> vwb_lens b \<Longrightarrow>a\<bowtie>b \<Longrightarrow> 
  a:[a :== \<guillemotleft>x\<guillemotright>; b :== \<guillemotleft>y\<guillemotright>] = a :== \<guillemotleft>x\<guillemotright>"
  unfolding lens_indep_def
  apply (simp add: prog_rep_eq)
  apply rel_simp
  apply (metis vwb_lens.put_eq)
  done

lemma 
  "vwb_lens a \<Longrightarrow> vwb_lens b \<Longrightarrow>
  a:[a :== \<guillemotleft>x\<guillemotright>; a:\<lbrakk>a :== \<guillemotleft>y\<guillemotright>\<rbrakk>] = a :== \<guillemotleft>x\<guillemotright>"
  unfolding lens_indep_def
  apply (simp add: prog_rep_eq)
  apply rel_simp
  apply (metis vwb_lens.put_eq)
  done

lemma antiframe_prog_idemp:
  "vwb_lens g \<Longrightarrow>g:[g:[P]] = g:[P]"
  apply (simp add: prog_rep_eq)
  apply rel_simp
  apply safe
   apply auto[1]
  apply (metis vwb_lens_weak weak_lens.put_get)
  done

lemma antiframe_prog_idemp_strong:
  assumes vwb_lens1: "vwb_lens g"
  assumes vwb_lens2: "vwb_lens g'"
  assumes region: "g' \<subseteq>\<^sub>L g"
  shows "g:[g':[P]] = g':[P]"
  apply (simp add: prog_rep_eq)
  apply rel_simp
  apply safe
   apply auto
  subgoal for  _ more _ x  
    apply (rule exI[where x ="x"])
    using region unfolding sublens_def
    apply (auto simp add: lens_get_put_quasi_commute lens_put_of_quotient vwb_lens1)
    done  
  apply (metis (no_types, lifting) lens_get_put_quasi_commute lens_put_of_quotient region vwb_lens1 vwb_lens_wb wb_lens.get_put)
  done

lemma antiframe_prog_monoI:  
  assumes "P1\<sqsubseteq>P2"
  shows "g:[P1]\<sqsubseteq> g:[P2]"
  apply (insert assms)  
  apply (simp add: prog_rep_eq)
  apply rel_blast
  done

lemma antiframe_prog_mono:
  "mono (pantiframe_prog G)"
  by (simp add: antiframe_prog_monoI monoI)

lemma Lower_prog_helper1:
  assumes  modifies: "\<forall>P. P \<in> (Lower_prog A) \<longrightarrow> Q P"
  shows "Q (\<Sqinter>\<^sub>pA)"
  apply (insert )  
  apply (simp add: assms inf_prog_in_lower_prog)
  done

lemma seq_prog_anti_frame_prog_distr_strong:
  assumes vwb_lens: "vwb_lens g"
  assumes modifies1: "P \<sqsubseteq> g:[P]" 
  shows "g:[P;Q] \<sqsubseteq> g:[P] ; g:[Q]"
  apply (insert assms)  
  apply (simp add: prog_rep_eq)
  apply rel_simp
  defer
  apply blast
  done

lemma seq_prog_anti_frame_prog_distr:
  assumes mwb_lens: "mwb_lens g"
  assumes modifies: "g:[P] = P"
  shows "(g:[P] ; g:[Q]) = (g:[P;Q])"
  apply (insert assms)  
  apply (simp add: prog_rep_eq)
  apply rel_auto
     apply transfer
     apply fastforce
    apply (metis (full_types) mwb_lens.put_put)   
   apply blast
  apply (metis des_vars.ext_inject mwb_lens.put_put mwb_lens )
  done

lemma seq_prog_anti_frame_prog_refines:
  assumes mwb_lens: "mwb_lens g"
  assumes modifies1: "g:[P1] = P1"
  assumes modifies2: "g:[P2] = P2"
  shows "g:[P1;P2] \<sqsubseteq> P1;P2"
  apply (insert assms)  
  apply (subst modifies1[THEN sym])
  apply (subst modifies2[THEN sym])
  apply (subst seq_prog_anti_frame_prog_distr[OF mwb_lens modifies1])
  apply (rule order.refl)
  done

lemma seq_prog_anti_frame_prog_refinedBy:
  assumes mwb_lens: "mwb_lens g"
  assumes modifies1: "g:[P1] = P1"
  assumes modifies2: "g:[P2] = P2"
  shows " P1;P2 \<sqsubseteq> g:[P1;P2] "
  apply (insert assms)  
  apply (subst (2) modifies1[THEN sym])
  apply (subst (2) modifies2[THEN sym])
  apply (subst seq_prog_anti_frame_prog_distr[OF mwb_lens modifies1])
  apply (rule order.refl)
  done

lemma seq_prog_anti_frame_prog_modifies:
  assumes mwb_lens: "mwb_lens g"
  assumes modifies1: "g:[P1] = P1"
  assumes modifies2: "g:[P2] = P2"
  shows "g:[P1;P2] = P1;P2"
  by (simp add: eq_iff modifies1 modifies2 seq_prog_anti_frame_prog_refinedBy 
      seq_prog_anti_frame_prog_refines mwb_lens)

lemma If_prog_anti_frame_prog_distr:
  "(IF b THEN g:[P1] ELSE g:[P2] FI) = (g:[IF b THEN P1 ELSE P2 FI])"
  apply (simp add: prog_rep_eq)
  apply rel_auto
  done

lemma If_prog_antiframe_prog_refines:
  assumes modifies1: "g:[P1] = P1"
  assumes modifies2: "g:[P2] = P2"
  shows "g:[IF b THEN P1 ELSE P2 FI] \<sqsubseteq> IF b THEN P1 ELSE P2 FI"
  apply (insert assms)  
  apply (subst modifies1[THEN sym])
  apply (subst modifies2[THEN sym])
  apply (subst If_prog_anti_frame_prog_distr)
  apply (rule order.refl)
  done

lemma If_prog_antiframe_prog_refinedBy_strong:
  assumes modifies1: "P1 \<sqsubseteq> g:[P1]"
  assumes modifies2: "P2  \<sqsubseteq> g:[P2] "
  shows "IF b THEN P1 ELSE P2 FI \<sqsubseteq> g:[IF b THEN P1 ELSE P2 FI] "
  apply (insert assms)  
  apply (subst If_prog_anti_frame_prog_distr[THEN sym])
  apply (simp add: if_prog_monoI)
  done

lemma If_prog_antiframe_prog_refinedBy:
  assumes modifies1: "g:[P1] = P1"
  assumes modifies2: "g:[P2] = P2"
  shows "IF b THEN P1 ELSE P2 FI \<sqsubseteq> g:[IF b THEN P1 ELSE P2 FI] "
  apply (insert assms)  
  apply (subst (2) modifies1[THEN sym])
  apply (subst (2) modifies2[THEN sym])
  apply (subst If_prog_anti_frame_prog_distr)
  apply (rule order.refl)
  done

lemma If_prog_anti_frame_prog_modifies:
  assumes modifies1: "g:[P1] = P1"
  assumes modifies2: "g:[P2] = P2"
  shows "g:[IF b THEN P1 ELSE P2 FI] = IF b THEN P1 ELSE P2 FI"
  by (metis If_prog_anti_frame_prog_distr modifies1 modifies2)

definition 
  "get_put' g \<sigma> = \<sigma>(g \<mapsto>\<^sub>s \<langle>\<sigma>\<rangle>\<^sub>s g)"

lemma
  "(put\<^bsub>g\<^esub> (\<sigma> s) (get\<^bsub>g\<^esub> (\<sigma> s)))  =  (\<sigma>(g \<mapsto>\<^sub>s \<langle>\<sigma>\<rangle>\<^sub>s g)) s"
  by rel_simp

lemma
  "(put\<^bsub>g\<^esub> s (get\<^bsub>g\<^esub> (\<sigma> s)))  =  (id(g \<mapsto>\<^sub>s \<langle>\<sigma>\<rangle>\<^sub>s g)) s"
  by rel_simp

lemma
  "(put\<^bsub>g\<^esub> s (get\<^bsub>g\<^esub> s))  =  (get_put' g id) s"
  unfolding get_put'_def
  by rel_simp 
term "(put\<^bsub>g\<^esub>  \<sigma> (get\<^bsub>g\<^esub>  \<sigma>)) = \<sigma>"
lemma assigns_prog_antiframe_prog_modifies:
  "[g \<mapsto>\<^sub>s \<langle>\<sigma>\<rangle>\<^sub>s g] = \<sigma> \<Longrightarrow>g:[\<langle>\<sigma>\<rangle>\<^sub>a] = \<langle>\<sigma>\<rangle>\<^sub>a"
  apply (simp add: prog_rep_eq)
  apply rel_simp
  done
lemma "(\<sigma>(g \<mapsto>\<^sub>s \<langle>\<sigma>\<rangle>\<^sub>s g) = \<sigma>) = wb_lens g"
  apply rel_auto
  using wb_lens.get_put[of g ] 
  apply simp
  
  oops
lemma "([g \<mapsto>\<^sub>s \<langle>\<sigma>\<rangle>\<^sub>s g] = \<sigma>) =  vwb_lens g"
  
  oops
lemma SKIP_prog_anti_frame_prog_modifies:
  "vwb_lens g \<Longrightarrow> g:[SKIP] = SKIP"
   apply (simp add: prog_rep_eq)
  apply rel_auto
  done

lemma antiframe_prog_refines_while_lfp_prog: 
  assumes vwb_lens: "vwb_lens g"
  assumes modifies: "g:[body] = body"
  shows "WHILE b DO body OD \<sqsubseteq> (g:[WHILE b DO body OD])"
  unfolding while_lfp_prog_mu_prog 
  apply (rule lfp_prog_lowerboundI)
  apply (fold while_lfp_prog_mu_prog)
  apply (subst (2) while_lfp_prog_unfold)
  apply (subst If_prog_anti_frame_prog_distr[THEN sym])
  apply (subst seq_prog_anti_frame_prog_distr[THEN sym])
  apply (simp add: assms)+
  apply (subst SKIP_prog_anti_frame_prog_modifies[OF vwb_lens])
   apply (rule refl)
  done

lemma antiframe_magic:  
  "a:[MAGIC] = MAGIC"
  apply (simp add: prog_rep_eq)
  apply rel_simp
  done

lemma antiframe_id_lens_skip: 
  "id_lens:[SKIP] = SKIP"
  apply (simp add: prog_rep_eq)
  apply rel_blast
  done

lemma antiframe_id_lens_assign:
  "id_lens:[\<langle>\<sigma>\<rangle>\<^sub>p] = \<langle>\<sigma>\<rangle>\<^sub>p"
  apply (simp add: prog_rep_eq)
  apply rel_blast
  done

lemma antiframe_zero_lens_abort:
  "zero_lens:[ABORT] = SKIP"
  apply (simp add: prog_rep_eq)
  apply rel_blast
  done

lemma antiframe_zero_lens_assign:
  "zero_lens:[\<langle>\<sigma>\<rangle>\<^sub>p] = SKIP"
  apply (simp add: prog_rep_eq)
  apply rel_auto
  done

lemma antiframe_zero_lens_skip:
  "zero_lens:[SKIP] = SKIP"
  apply (simp add: prog_rep_eq)
  apply rel_auto
  done

subsection \<open>laws for frame \<close>

lemma frame_test:
  "vwb_lens a \<Longrightarrow> vwb_lens b \<Longrightarrow> a\<bowtie>b \<Longrightarrow> a:\<lbrakk>a :== \<guillemotleft>x\<guillemotright>; b :== \<guillemotleft>y\<guillemotright>\<rbrakk> = b:== \<guillemotleft>y\<guillemotright>"
  unfolding lens_indep_def
  apply (simp add: prog_rep_eq)
  apply rel_simp
  apply (metis vwb_lens.put_eq)
  done

lemma frame_prog_idemp:
  "vwb_lens g \<Longrightarrow>g:\<lbrakk>g:\<lbrakk>P\<rbrakk>\<rbrakk> = g:\<lbrakk>P\<rbrakk>"
  apply (simp add: prog_rep_eq)
  apply rel_auto
  done

lemma frame_prog_monoI:  
  assumes "P1\<sqsubseteq>P2"
  shows "g:\<lbrakk>P1\<rbrakk>\<sqsubseteq> g:\<lbrakk>P2\<rbrakk>"
  apply (insert assms)  
  apply (simp add: prog_rep_eq)
  apply rel_blast
  done

lemma frame_prog_mono:
  "mono (pframe_prog G)"
  by (simp add: frame_prog_monoI monoI)

lemma frame_zero_lens_magic:
  "a:\<lbrakk>MAGIC\<rbrakk> = MAGIC"
  apply (simp add: prog_rep_eq)
  apply rel_auto
  done

lemma frame_zero_lens_assign:
  "zero_lens:\<lbrakk>\<langle>\<sigma>\<rangle>\<^sub>p\<rbrakk> = \<langle>\<sigma>\<rangle>\<^sub>p"
  apply (simp add: prog_rep_eq)
  apply rel_auto
  done

lemma frame_id_lens_assign:
  "id_lens:\<lbrakk>\<langle>\<sigma>\<rangle>\<^sub>p\<rbrakk> = SKIP"
  apply (simp add: prog_rep_eq)
  apply rel_auto
  done

section \<open>General laws on lattices\<close>

lemma inf_prog_Int_abrh:
  "(\<And>x. x \<in> A \<Longrightarrow> abrh x = x) \<Longrightarrow> 
     \<Sqinter>\<^sub>p(A \<inter> {P. abrh P = P}) = \<Sqinter>\<^sub>p A"
  apply (simp add: prog_rep_eq)
  unfolding inf_def image_def Ball_def 
  apply (rule someI2_ex)
   apply (rule exI[where x="(\<^bold>\<Sqinter>\<^bsub>NDES\<^esub>{y. \<exists>x\<in>A. y = \<lbrakk>x\<rbrakk>\<^sub>p})"])
   apply (rule normal_design_theory_continuous.inf_glb)
  using Rep_prog apply auto[]
  apply (rule ndes_utp_theory.greatest_unique)
   defer
   apply auto[1]
  apply (rule someI2_ex)
   apply (rule exI[where x="(\<^bold>\<Sqinter>\<^bsub>NDES\<^esub>{y. \<exists>x\<in>A \<inter> {P. \<lbrakk>abrh P\<rbrakk>\<^sub>p = \<lbrakk>P\<rbrakk>\<^sub>p}. y = \<lbrakk>x\<rbrakk>\<^sub>p})"])
   apply (rule normal_design_theory_continuous.inf_glb)
  using Rep_prog apply auto[]
  unfolding Bex_def
  apply (subst (asm) Int_Collect)
  apply simp
  apply (subst Collect_cong)
   apply simp_all
  apply blast
  done

lemma inf_prog_abrh_greatest_prog:
  "(\<And>x. x \<in> A \<Longrightarrow> abrh z \<sqsubseteq> abrh x) \<Longrightarrow> abrh z \<sqsubseteq> \<Sqinter>\<^sub>p (abrh ` A)"
  by (metis (mono_tags, lifting) imageE inf_prog_greatest)

lemma inf_prog_abrh_image_lower_prog:
  "x \<in> A \<Longrightarrow> \<Sqinter>\<^sub>p (abrh ` A) \<sqsubseteq> abrh x"
  by (simp add: inf_prog_lower)   

lemma inf_prog_abrh_in_Lower_prog_Int_abrh:
  "mono abrh \<Longrightarrow>abrh (\<Sqinter>\<^sub>pA) \<in> Lower_prog (A \<inter> {P. abrh P = P})" 
  unfolding inf_prog_alt_def 
  apply (rule someI2_ex)    
  using inf_prog_is_greatest_lower_prog 
   apply blast
  apply (simp only: greatest_prog_alt_def Lower_prog_alt_def)
  apply auto
  using monoD 
  apply fastforce
  done

section \<open>laws for scoping and Fixed point\<close>

abbreviation
  "mu_scp G F \<equiv> CONST prec_lfp_prog (CONST comp F (CONST pantiframe_prog G))"

syntax
  "_smu" :: "logic \<Rightarrow> pttrn \<Rightarrow> logic \<Rightarrow> logic" ("\<mu>\<^bsub>_\<^esub> _ \<bullet> _" [0,0,10] 10)

notation mu_scp ("\<mu>\<^bsub>_\<^esub>")

translations 
  "\<mu>\<^bsub>G\<^esub> X \<bullet> P" == "CONST prec_lfp_prog (CONST comp (\<lambda> X. P) (CONST pantiframe_prog G))" 

lemma mu_scp_antiframe_refinedBy:
  assumes M: "mono_prog F"
  assumes H : "(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G':[P] = P}"
  shows "G':[\<mu>\<^bsub>G\<^esub> F] \<sqsubseteq> (\<mu>\<^bsub>G\<^esub> F)"
  unfolding lfp_prog_alt_def
proof (auto)
  have a1: "\<forall>i. G':[F (G:[i])] = F (G:[i])"
    using H 
    by auto
  then have "mono_prog (\<lambda>p. F (G:[p]))"
    using M
    by (meson Mono_progE Mono_progI antiframe_prog_monoI)
  then show "G':[\<Sqinter>\<^sub>p{p. F (G:[p]) \<sqsubseteq> p}] \<sqsubseteq> \<Sqinter>\<^sub>p{p. F(G:[p]) \<sqsubseteq> p}"
    apply (fold lfp_prog_alt_def)
    apply (subst (2) lfp_prog_unfold )
     apply assumption
    apply (subst lfp_prog_unfold)
     apply assumption
    apply (simp add: a1)
    done
qed

lemma mu_scp_antiframe_refines:
  assumes M: "mono_prog F"
  assumes H : "(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G':[P] = P}"  
  shows " \<mu>\<^bsub>G\<^esub> F  \<sqsubseteq> G':[\<mu>\<^bsub>G\<^esub> F] "
  unfolding lfp_prog_alt_def
proof (auto)
  have a1: "\<forall>i. G':[F (G:[i])] = F (G:[i])"
    using H by auto
  then have "mono_prog (\<lambda>p. F (G:[p]))"
    using M
    by (meson Mono_progE Mono_progI antiframe_prog_monoI)
  then show " \<Sqinter>\<^sub>p{p.  F (G:[p]) \<sqsubseteq> p}  \<sqsubseteq> G':[\<Sqinter>\<^sub>p{x. F (G:[x]) \<sqsubseteq> x}] "
    apply (fold lfp_prog_alt_def)
    apply (subst a1[THEN spec, THEN sym])
    apply (subst (2)lfp_prog_unfold)
     apply assumption
    apply (subst (2)a1[THEN spec ,THEN sym])
    apply (subst lfp_prog_unfold)
     apply (simp_all add: a1 )
    done
qed

lemma mu_scp_antiframe_modifies[closure]:
  assumes M: "mono_prog F"
  assumes H : "(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G':[P] = P}"
  shows "G':[\<mu>\<^bsub>G\<^esub> F] = (\<mu>\<^bsub>G\<^esub> F)"
  apply (rule antisym)
   apply (simp_all add: assms mu_scp_antiframe_refinedBy mu_scp_antiframe_refines)
  done

lemma modifies_indep':
  "vwb_lens G \<Longrightarrow> vwb_lens l \<Longrightarrow> G \<bowtie> l \<Longrightarrow> G:[P] = P \<Longrightarrow> {&G, &l}:[P] = G:[P]"
  unfolding pr_var_def
  apply (simp add: prog_rep_eq)
  apply rel_auto
  unfolding lens_indep_def
   apply simp 
   apply (metis lens_override_def lens_override_idem)+
  done

lemma modifies_indep:
  "vwb_lens G \<Longrightarrow> vwb_lens l \<Longrightarrow> G \<bowtie> l \<Longrightarrow> G:[P] = P \<Longrightarrow> MODIFY G +\<^sub>L l DO P OD = P"
  unfolding pr_var_def
  apply (simp add: prog_rep_eq)
  apply rel_auto
  unfolding lens_indep_def
   apply simp 
   apply (metis lens_override_def lens_override_idem)+
  done

lemma (*USE THIS*)
  "G:[body] = body \<Longrightarrow> 
   ((\<lambda>X. IF b THEN body ; X ELSE SKIP FI)o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G:[P] = P}"
  apply auto
  oops

lemma lfp_scp_prog_lowerboundI_weak:
  "G:[x] = x \<Longrightarrow> F x \<sqsubseteq> x \<Longrightarrow> \<mu>\<^bsub>G\<^esub> F \<sqsubseteq> x"
  unfolding lfp_prog_alt_def  
  by (auto intro!: inf_prog_lower)

lemma lfp_scp_prog_lowerboundI_strong:
  "F (G:[x]) \<sqsubseteq> x \<Longrightarrow> \<mu>\<^bsub>G\<^esub> F \<sqsubseteq> x"
  unfolding lfp_prog_alt_def  
  by (auto intro!: inf_prog_lower)

lemma lfp_scp_prog_greatest_strong:
  "(\<And>z. F (G:[z]) \<sqsubseteq> z \<Longrightarrow> x \<sqsubseteq> z) \<Longrightarrow> x \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F"  
  unfolding lfp_prog_alt_def  
  by (auto intro!: inf_prog_greatest)

lemma lfp_scp_prog_fixed_point_strong:
  assumes M:"mono_prog F"
  assumes H: "(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G:[P] = P}"
  shows "F (\<mu>\<^bsub>G\<^esub> F) = \<mu>\<^bsub>G\<^esub> F" 
proof (rule antisym)
  have ne: "{P.  F(G:[P]) \<sqsubseteq> P} \<noteq> {}"
  proof -
    have "F (G:[MAGIC]) \<sqsubseteq> (G:[MAGIC])"
      by (simp add: antiframe_magic magic_prog_refines_any)
    thus ?thesis       
      by (auto, rule_tac x="(G:[MAGIC])" in exI, 
          auto simp: antiframe_magic)
  qed
  show "\<mu>\<^bsub>G\<^esub> F \<sqsubseteq> F (\<mu>\<^bsub>G\<^esub> F)"
  proof -
    have "mono_prog (F \<circ> pantiframe_prog G)"
      by (metis (no_types) M Mono_progE Mono_progI antiframe_prog_mono comp_apply mono_def)
    then show ?thesis
      using eq_refl lfp_prog_fixed_point
      by (metis (mono_tags) H M comp_apply mu_scp_antiframe_modifies)
  qed
  from ne show "F (\<mu>\<^bsub>G\<^esub>F) \<sqsubseteq> \<mu>\<^bsub>G\<^esub>F"
  proof -
    have "mono_prog (F \<circ> pantiframe_prog G)"
      by (metis (no_types) M Mono_progE Mono_progI antiframe_prog_mono comp_apply mono_def)
    then show ?thesis
      using eq_refl lfp_prog_unfold      
      by (metis (mono_tags) H M comp_apply mu_scp_antiframe_modifies)
  qed
qed

lemma lfp_scp_prog_fixed_point_weak:
  assumes M:"mono_prog F"
  assumes H: "(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G:[P] = P}"
  shows "(F o pantiframe_prog G) (\<mu>\<^bsub>G\<^esub> F) = \<mu>\<^bsub>G\<^esub> F"   
  by (simp add: H M lfp_scp_prog_fixed_point_strong mu_scp_antiframe_modifies)

lemma lfp_scp_prog_unfold_strong: 
  assumes M:"mono_prog F"
  assumes H: "(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G:[P] = P}"
  shows "(\<mu>\<^bsub>G\<^esub> F) = F(\<mu>\<^bsub>G\<^esub> F)"
  by (rule lfp_scp_prog_fixed_point_strong[THEN sym, OF assms])

lemma lfp_scp_prog_unfold_weak: 
  assumes M:"mono_prog F"
  assumes H: "(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G:[P] = P}"
  shows "(\<mu>\<^bsub>G\<^esub> F) = (F o pantiframe_prog G) (\<mu>\<^bsub>G\<^esub> F)"
  by (rule lfp_scp_prog_fixed_point_weak[THEN sym, OF assms])

lemma lfp_scp_prog_const:
  "\<mu>\<^bsub>G\<^esub> (\<lambda>x. t) = t"
  apply (subst lfp_prog_unfold) 
   apply (auto simp: comp_def Mono_progI)
  done

lemma antiframe_prog_is_idemp:
  "vwb_lens G \<Longrightarrow>idem_prog (pantiframe_prog G)"
  apply (simp add: prog_rep_eq)
  unfolding utp_order_def ndes_hcond_def idempotent_def Ball_def
  apply auto
  apply (subst Rep_prog_inverse)
  apply (subst antiframe_prog_idemp)
   apply (assumption)
  apply (rule refl)
  done

lemma lfp_scp_prog_idem:
  "mono_prog F \<Longrightarrow> idem_prog (F o pantiframe_prog G) \<Longrightarrow> \<mu>\<^bsub>G\<^esub> F = (F o pantiframe_prog G) ABORT"
proof (subst lfp_prog_idem, goal_cases)
  case 1
  then show ?case 
    by (metis Mono_progE Mono_progI antiframe_prog_mono comp_def mono_def)  
next
  case 2
  then show ?case by simp
next
  case 3
  then show ?case by simp
qed

lemma lfp_scp_prog_eqI:
  assumes M:"mono_prog F"
  assumes H: "(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G:[P] = P}"
  assumes L: "F(G:[x]) \<sqsubseteq> x"
  assumes G: "(\<And>z. F (G:[z]) = z \<Longrightarrow> x \<sqsubseteq> z )"
  shows "\<mu>\<^bsub>G\<^esub>F = x"
  apply (rule antisym)
   apply (simp_all add: assms lfp_scp_prog_lowerboundI_strong )  
proof -
  have "F (G:[\<mu>\<^bsub>G\<^esub> F]) = \<mu>\<^bsub>G\<^esub> F" 
    apply (subst lfp_scp_prog_fixed_point_strong[unfolded comp_def])
      apply (simp add: Mono_progI)
    using H apply auto[]
    apply (subst (2) lfp_scp_prog_unfold_weak[OF M H])
    apply auto
    apply (subst lfp_prog_const)
    apply (rule refl)
    done
  then show "x \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F"
    by (metis G)
qed

lemma lfp_scp_prog_ordinal_induct[case_names M H step union]:
  assumes M:"mono_prog F"
  assumes H: "(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G:[P] = P}"
  assumes P_f:   "\<And>S. P S \<Longrightarrow> S \<sqsubseteq> \<mu>\<^bsub>G\<^esub>F \<Longrightarrow> P ((F o pantiframe_prog G) S)"
  assumes P_Union: "\<And>M. (\<And>S. S \<in> M \<Longrightarrow>  P S) \<Longrightarrow> P (sup_prog M)"
  shows "P (\<mu>\<^bsub>G\<^esub>F)"
proof -
  let ?M = "{S. S \<sqsubseteq> \<mu>\<^bsub>G\<^esub>F \<and> P S}"
  from P_Union have "P (sup_prog ?M)" by (smt Collect_mono_iff mem_Collect_eq)
  also have "sup_prog ?M = \<mu>\<^bsub>G\<^esub>F"
  proof (rule antisym)
    show "sup_prog ?M \<sqsubseteq> \<mu>\<^bsub>G\<^esub>F"
      apply (subst sup_prog_least)
       apply (auto simp add: Collect_mono)
      done
    then have "(F o pantiframe_prog G) (sup_prog ?M)  \<sqsubseteq> F (\<mu>\<^bsub>G\<^esub>F)"
    proof -
      have "\<forall>f. f (\<Squnion>\<^sub>p {p. p \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F \<and> P p}) \<sqsubseteq> f (\<mu>\<^bsub>G\<^esub> F) \<or> \<not> mono_prog f"
        by (metis Mono_progE \<open>\<Squnion>\<^sub>p{S. S \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F \<and> P S} \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F\<close>) 
      then show ?thesis
        by (metis (mono_tags) H M Mono_progE antiframe_prog_mono comp_apply mono_Monotone_prog mu_scp_antiframe_modifies)
    qed
    then have "(F o pantiframe_prog G) (sup_prog ?M) \<sqsubseteq> \<mu>\<^bsub>G\<^esub>F"
      apply (subst lfp_prog_unfold)
       apply (simp_all add: H M mu_scp_antiframe_modifies Mono_progD Mono_progI antiframe_prog_monoI)
      done
    then have "(F o pantiframe_prog G) (sup_prog ?M) \<in> ?M"
      using P_Union  P_f \<open>\<Squnion>\<^sub>p{S. S \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F \<and> P S} \<sqsubseteq> \<mu>\<^bsub>G\<^esub> F\<close> calculation 
      by auto
    then have "(F o pantiframe_prog G) (sup_prog ?M)  \<sqsubseteq> sup_prog ?M"
      by (simp add: Collect_mono sup_prog_upper)
    then show "\<mu>\<^bsub>G\<^esub>F \<sqsubseteq> sup_prog ?M"
      apply (subst lfp_scp_prog_lowerboundI_strong)
       apply (simp_all add: Collect_mono lfp_scp_prog_lowerboundI_strong assms )
      done
  qed
  finally show ?thesis.
qed

theorem lfp_scp_prog_induct:
  assumes M:"mono_prog F"
  assumes H: "(F o pantiframe_prog G) \<in> UNIV \<rightarrow> {P. G:[P] = P}"
  assumes ind: "F (join_prog  (\<mu>\<^bsub>G\<^esub>F) P) \<sqsubseteq> P"
  shows "\<mu>\<^bsub>G\<^esub>F \<sqsubseteq> P"
proof (induct rule: lfp_scp_prog_ordinal_induct)
  case M
  then show ?case
    by fact
next
  case H
  then show ?case by fact
next
  case (step S)
  then show ?case
    apply (intro order_trans[OF ind _])
    apply auto
    apply (insert M)
    apply (erule Mono_progD)
    apply (rule le_join_progI1)
    apply (subst  mu_scp_antiframe_modifies[where G' = G,THEN sym])
      apply (simp_all add: assms)
    apply (rule antiframe_prog_monoI)
    apply assumption
    done
next
  case (union M)
  then show ?case  by (auto intro!: sup_prog_least)
qed

section \<open>laws for iterations and scoping\<close>

definition from_until_lfp_scp_prog ::
  " ('b \<Longrightarrow>'\<alpha>)\<Rightarrow>'\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ FROM (_)/ UNTIL (_)/ DO (_) OD")
  where "IN G FROM init UNTIL exit DO body OD \<equiv> 
         init ; (\<mu>\<^bsub>G\<^esub> X \<bullet> IF \<not> exit THEN (body ; X) ELSE SKIP FI)"

declare from_until_lfp_scp_prog_def[prog_rep_eq]

definition from_until_lfp_scp_invr_prog :: 
  "('b \<Longrightarrow>'\<alpha>)\<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ FROM (_)/ INVAR (_)/ UNTIL (_)/  DO (_) OD")  
  where "IN G  FROM init INVAR invar UNTIL exit DO body OD = IN G FROM init UNTIL exit DO body OD"

declare from_until_lfp_scp_invr_prog_def[prog_rep_eq]

definition from_until_lfp_scp_invr_vrt_prog :: 
  "('b \<Longrightarrow>'\<alpha>)\<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ FROM (_)/ INVAR (_)/ VRT (_)/ UNTIL (_)/ DO (_) OD")
  where "IN G FROM init INVAR invar VRT vari UNTIL exit DO body OD = IN G FROM init UNTIL exit DO body OD" 

declare from_until_lfp_scp_invr_vrt_prog_def[prog_rep_eq]

definition while_lfp_scp_prog :: 
  "('b \<Longrightarrow>'\<alpha>) \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ WHILE (_)/ DO (_) OD")
  where "IN G WHILE b DO body OD = IN G FROM SKIP UNTIL \<not>b DO body OD"   

declare while_lfp_scp_prog_def[prog_rep_eq]

definition while_lfp_scp_invr_prog :: 
  "('b \<Longrightarrow>'\<alpha>) \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ INVAR (_)/ WHILE (_)/ DO (_) OD")
  where "IN G INVAR invar WHILE b DO body OD = IN G WHILE b DO body OD" 

declare while_lfp_scp_invr_prog_def[prog_rep_eq]

definition while_lfp_scp_invr_vrt_prog :: 
  "('b \<Longrightarrow>'\<alpha>) \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ INVAR (_)/ VRT (_)/ WHILE (_)/ DO (_) OD")
  where "IN G INVAR invar VRT vari WHILE b DO body OD = IN G WHILE b DO body OD"

declare while_lfp_scp_invr_vrt_prog_def[prog_rep_eq]

definition do_while_lfp_scp_prog :: 
  "('b \<Longrightarrow>'\<alpha>) \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ DO (_)/  WHILE (_) OD")
  where "IN G DO body WHILE b OD = IN G FROM body UNTIL \<not> b DO body OD"

declare do_while_lfp_scp_prog_def [prog_rep_eq]

definition do_while_lfp_scp_invr_prog :: 
  "('b \<Longrightarrow>'\<alpha>) \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ DO (_)/ INVAR (_)/ WHILE (_) OD")
  where "IN G DO body INVAR invar WHILE b OD = IN G DO body WHILE b OD"

declare do_while_lfp_scp_invr_prog_def [prog_rep_eq]

definition do_while_lfp_scp_invr_vrt_prog :: 
  "('b \<Longrightarrow>'\<alpha>) \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ DO (_)/ INVAR (_)/ VRT (_)/ WHILE (_) OD")
  where "IN G DO body INVAR invar VRT vari WHILE b OD = IN G DO body WHILE b OD"  

declare do_while_lfp_scp_invr_vrt_prog_def [prog_rep_eq]

definition for_lfp_scp_prog :: 
  "('b \<Longrightarrow>'\<alpha>) \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ FOR ('(_,_,_'))/ DO (_) OD")
  where "IN G FOR (init, b, incr) DO body OD = IN G FROM init UNTIL \<not> b DO body ; incr OD"

declare for_lfp_scp_prog_def [prog_rep_eq]

definition for_lfp_scp_invr_prog :: 
  "('b \<Longrightarrow>'\<alpha>) \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ FOR ('(_,_,_'))/ INVAR (_)/ DO (_) OD")
  where "IN G FOR (init, b, incr) INVAR invar DO body OD = IN G FOR (init, b, incr) DO body OD"  

declare for_lfp_scp_invr_prog_def [prog_rep_eq]

definition for_lfp_scp_invr_vrt_prog :: 
  "('b \<Longrightarrow>'\<alpha>) \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> cond \<Rightarrow> ('t,'\<alpha>) uexpr \<Rightarrow> '\<alpha> prog \<Rightarrow> '\<alpha> prog" 
  ("IN (_)/ FOR ('(_,_,_'))/ INVAR (_)/ VRT (_)/ DO (_) OD")
  where "IN G FOR (init, b, incr) INVAR invar VRT vari DO body OD = IN G FOR (init, b, incr) DO body OD"  

declare for_lfp_scp_invr_vrt_prog_def [prog_rep_eq]

text \<open>We introduce the algebraic laws of programming related to the iterations combined with
      the scoping operator.\<close>

lemma while_lfp_scp_prog_mu_prog:
  "IN G WHILE b DO body OD = (\<mu>\<^bsub>G\<^esub> X \<bullet> IF b THEN (body ; X) ELSE SKIP FI)"
  unfolding  while_lfp_scp_prog_def from_until_lfp_scp_prog_def
  by simp

theorem while_lfp_scp_prog_unfold:
  assumes vwb_lens: "vwb_lens G"
  assumes modifies: "G:[body] = body"
  shows "IN G WHILE b DO body OD = (IF b THEN (body ; IN G WHILE b DO body OD) ELSE SKIP FI)"
proof -  
  have "(IN G WHILE b DO body OD) = (\<mu>\<^bsub>G\<^esub> X \<bullet> IF b THEN body ; X ELSE SKIP FI)"
    by (simp add: while_lfp_scp_prog_mu_prog)
  also have "... = (IF b THEN (body ; (\<mu>\<^bsub>G\<^esub> X \<bullet> IF b THEN body ; X ELSE SKIP FI)) ELSE SKIP FI)"
  proof (rule lfp_scp_prog_unfold_strong[OF Mono_progI ], goal_cases)
    case (1 x y)
    then show ?case by (simp add: if_prog_monoI seq_prog_monoI)
  next
    case 2
    then show ?case      
      apply auto[]
      by (simp add: If_prog_anti_frame_prog_modifies SKIP_prog_anti_frame_prog_modifies 
          antiframe_prog_idemp modifies seq_prog_anti_frame_prog_modifies vwb_lens)
  qed  
  also have "... = (IF b THEN (body ; IN G WHILE b DO body OD) ELSE SKIP FI)"
    unfolding while_lfp_scp_prog_mu_prog
    by (rule refl)
  finally show ?thesis .
qed

theorem while_lfp_scp_prog_true: 
  "IN G WHILE true DO body OD = (\<mu>\<^bsub>G\<^esub> X \<bullet>  body ; X)" 
  unfolding while_lfp_scp_prog_mu_prog  if_prog_true ..
 
lemma while_lfp_scp_prog_false:
  "(IN G WHILE false DO body OD) = SKIP"
  unfolding while_lfp_scp_prog_mu_prog if_prog_false
  by (simp add: lfp_scp_prog_const) 

lemma while_lfp_scp_prog_non_termination:
  "vwb_lens G \<Longrightarrow>IN G WHILE true DO SKIP OD = G:[ABORT]"
  unfolding while_lfp_scp_prog_mu_prog if_prog_true 
proof -
  have 1: "(op ; SKIP) = (\<lambda>x. SKIP ; x)"
    by auto
  also have 2: "G:[ABORT] = (op ; SKIP \<circ> pantiframe_prog G) ABORT"
    by simp
  show "vwb_lens G \<Longrightarrow> \<mu>\<^bsub>G\<^esub> (op ; SKIP) = G:[ABORT]"
    apply (subst 1)
    apply (subst 2)
    apply (rule lfp_scp_prog_idem)
     apply (simp add: Mono_progI)
    apply (simp add: prog_rep_eq)
    unfolding utp_order_def ndes_hcond_def idempotent_def Ball_def
    apply auto
    apply (subst Rep_prog_inverse)
    apply (subst antiframe_prog_idemp)
     apply (assumption)
    apply (rule refl)
    done
qed

lemma from_until_lfp_scp_prog_def_alt:
  "IN G FROM init UNTIL b DO body OD = init ; IN G WHILE \<not>b DO body OD"
  unfolding from_until_lfp_scp_prog_def while_lfp_scp_prog_mu_prog ..

lemma from_until_lfp_scp_invr_vrt_prog_def_alt:
  "IN G FROM init INVAR invar VRT vari UNTIL b DO body OD = 
   init ; IN G INVAR invar VRT vari WHILE \<not>b DO body OD"
  by (simp add: from_until_lfp_scp_invr_vrt_prog_def from_until_lfp_scp_prog_def_alt 
                while_lfp_scp_invr_vrt_prog_def)

lemma from_until_lfp_scp_prog_unfold:
  "vwb_lens G \<Longrightarrow> G:[body] = body \<Longrightarrow> 
  IN G FROM init UNTIL b DO body OD = init ; IF \<not>b THEN body ; IN G WHILE \<not>b DO body OD ELSE SKIP FI"
  unfolding from_until_lfp_scp_prog_def_alt 
  using while_lfp_scp_prog_unfold[of G body "\<not>b"]
  by simp

lemma from_until_lfp_scp_prog_true: 
  "IN G FROM init UNTIL true DO body OD = init"
  unfolding from_until_lfp_scp_prog_def_alt
  by (simp add: while_lfp_scp_prog_false)

lemma from_until_lfp_scp_prog_non_termination: 
  "vwb_lens G \<Longrightarrow> IN G FROM init UNTIL false DO SKIP OD = init ; G:[ABORT]"
  unfolding from_until_lfp_scp_prog_def_alt
  by (simp add: while_lfp_scp_prog_non_termination)

lemma do_while_lfp_scp_prog_def_alt:
  "IN G DO body WHILE b OD = body ; IN G WHILE b DO body OD"
  unfolding do_while_lfp_scp_prog_def from_until_lfp_scp_prog_def_alt
  by simp

lemma do_while_lfp_scp_invr_vrt_prog_def_alt:
  "IN G DO body INVAR invar VRT vari WHILE b OD = 
   body ; IN G INVAR invar VRT vari WHILE b DO body OD"
  by (simp add: do_while_lfp_scp_invr_vrt_prog_def do_while_lfp_scp_prog_def_alt 
                while_lfp_scp_invr_vrt_prog_def)

lemma do_while_lfp_scp_prog_unfold: 
  "vwb_lens G \<Longrightarrow> G:[body] = body \<Longrightarrow> 
   IN G DO body WHILE b OD = body ; IF b THEN body; IN G WHILE b DO body OD ELSE SKIP FI"
  unfolding do_while_lfp_scp_prog_def_alt
  using while_lfp_scp_prog_unfold[of G  body b]
  by simp

lemma do_while_lfp_scp_prog_false:
  "IN G DO body WHILE false OD = body"
  unfolding do_while_lfp_scp_prog_def_alt
  by (simp add: while_lfp_scp_prog_false)

lemma do_while_lfp_scp_prog_non_termination:
  "vwb_lens G \<Longrightarrow> IN G DO SKIP WHILE true OD = G:[ABORT]" 
  unfolding do_while_lfp_scp_prog_def_alt
  by (simp add: while_lfp_scp_prog_non_termination)

lemma for_lfp_scp_prog_def_alt:
  "IN G FOR (init, b, incr) DO body OD = 
   init ; IN G WHILE b DO body ; incr OD"
  unfolding for_lfp_scp_prog_def from_until_lfp_scp_prog_def_alt
  by simp

lemma for_lfp_scp_invr_vrt_prog_def_alt:
  "IN G FOR (init, b, incr) INVAR invar VRT vari DO body OD = 
   init ; IN G INVAR invar VRT vari WHILE b DO body ; incr OD"
  by (simp add: for_lfp_scp_invr_vrt_prog_def for_lfp_scp_prog_def_alt 
                while_lfp_scp_invr_vrt_prog_def)

lemma for_lfp_scp_prog_unfold: 
  " vwb_lens G \<Longrightarrow> G:[body; incr] = body; incr \<Longrightarrow>
   IN G FOR (init, b, incr) DO body OD = 
   init; IF b THEN body; incr; IN G WHILE b DO body; incr OD ELSE SKIP FI"
  unfolding for_lfp_scp_prog_def_alt 
  using while_lfp_scp_prog_unfold [of G "body;incr" b ] seq_prog_assoc[THEN HOL.sym]
  by metis

lemma for_lfp_scp_prog_false:
  "IN G FOR (init, false, incr) DO body OD = init"
  unfolding for_lfp_scp_prog_def_alt
  by (simp add: while_lfp_scp_prog_false)

lemma for_lfp_scp_prog_non_termination:
  "IN G FOR (init, true, incr) DO SKIP OD = init; IN G WHILE true DO incr OD"
  unfolding for_lfp_scp_prog_def_alt
  by simp

lemma antiframe_prog_refines_while_lfp_scp_prog: 
  assumes vwb_lens: "vwb_lens G"
  assumes modifies: "G:[body] = body"
  shows "IN G WHILE b DO body OD \<sqsubseteq> G:[IN G WHILE b DO body OD]"
  unfolding while_lfp_scp_prog_mu_prog 
  apply (rule lfp_scp_prog_lowerboundI_strong)
  apply (subst antiframe_prog_idemp[OF vwb_lens])
  apply (fold while_lfp_scp_prog_mu_prog)
  apply (subst while_lfp_scp_prog_unfold[OF vwb_lens modifies])
  apply (subst If_prog_anti_frame_prog_distr[THEN sym])
  apply (subst SKIP_prog_anti_frame_prog_modifies[OF vwb_lens])
  apply (subst seq_prog_anti_frame_prog_distr[THEN sym])
    apply (simp add: assms)
  apply (simp add: assms)
  apply (subst modifies)
  apply (rule order.refl)
  done

lemma antiframe_prog_refinedBy_while_lfp_scp_prog:
  assumes vwb_lens: "vwb_lens G"
  assumes modifies: "G:[body] = body"
  shows   "G:[IN G WHILE b DO body OD] \<sqsubseteq> IN G WHILE b DO body OD"
  apply (subst (2) while_lfp_scp_prog_unfold[OF vwb_lens modifies])
  apply (subst If_prog_anti_frame_prog_distr[THEN sym])
  apply (subst SKIP_prog_anti_frame_prog_modifies[OF vwb_lens])
  apply (subst seq_prog_anti_frame_prog_distr[THEN sym])
    apply (simp add: assms)
  apply (simp add: assms)
  apply (subst modifies)
  unfolding while_lfp_scp_prog_mu_prog 
  apply (rule lfp_scp_prog_greatest_strong)
  apply (meson antiframe_prog_monoI if_prog_monoI lfp_scp_prog_lowerboundI_strong 
      order_refl order_trans seq_prog_monoI)
  done

lemma antiframe_prog_modifies_while_lfp_scp_prog:
  assumes vwb_lens: "vwb_lens G"
  assumes modifies: "G:[body] = body"
  shows   "G:[IN G WHILE b DO body OD] = IN G WHILE b DO body OD"
  by (simp add: antiframe_prog_refinedBy_while_lfp_scp_prog 
      antiframe_prog_refines_while_lfp_scp_prog antisym modifies vwb_lens)

find_theorems "\<Sqinter>\<^sub>p_ \<in> _"
thm antiframe_prog_mono
    antiframe_prog_monoI

lemma
  assumes \<open>\<Sqinter>\<^sub>pA \<in> A\<close>
  shows \<open>G:[\<Sqinter>\<^sub>pA] \<in> pantiframe_prog G ` A\<close>
  unfolding image_def Bex_def
  apply simp
  apply (rule exI[where x = \<open>\<Sqinter>\<^sub>pA\<close>])
  using assms
  apply simp
  done

lemma least_in_same_set_mono:
  \<open>mono F \<Longrightarrow> \<Sqinter>\<^sub>p{x.  F x \<sqsubseteq> x} \<in> {x.  F x \<sqsubseteq> x}\<close>
  apply auto
  apply (rule inf_prog_greatest)
  apply (metis (full_types) inf_prog_lower lfp_prog_alt_def lfp_prog_unfold mono_Monotone_prog)
  done

lemma anti_frame_is_continues:
  assumes least_in: \<open> \<Sqinter>\<^sub>pA \<in> A\<close>
  shows  \<open>pantiframe_prog G (\<Sqinter>\<^sub>p A) = \<Sqinter>\<^sub>p (image (pantiframe_prog G) A)\<close>
proof (rule antisym, goal_cases)
  case 1
  then show ?case
    apply (rule inf_prog_lower)
    apply (rule imageI)
    using least_in
     apply simp
    done
next
  case 2
  then show ?case
    apply (rule inf_prog_greatest)
    apply (erule imageE)
    apply simp
    apply (rule antiframe_prog_monoI)
    apply (rule inf_prog_lower) 
    apply auto
    done
qed

lemma \<open>mono F \<Longrightarrow> pantiframe_prog G (\<Sqinter>\<^sub>p{x.  F x \<sqsubseteq> x}) = P\<close>
  apply (subst anti_frame_is_continues)
  using least_in_same_set_mono apply blast
  oops
lemma \<open>mono F \<Longrightarrow>
   G:[\<mu>\<^sub>p ((pantiframe_prog G) o F)] = (\<mu>\<^sub>p ((pantiframe_prog G) o F))\<close>
  unfolding lfp_prog_alt_def
  apply simp
  apply (subst anti_frame_is_continues)
   apply simp
   apply (rule inf_prog_greatest)
   apply simp
   apply (rule order_trans)
    apply simp
   apply (rule antiframe_prog_monoI)
   apply (simp only: mono_def)
   apply (simp add: inf_prog_lower)
  unfolding image_def Bex_def
  apply simp
  apply (smt antiframe_prog_mono inf_prog_lower mem_Collect_eq mono_def order_subst1)
  thm antiframe_prog_monoI[THEN order_trans]
  thm order_trans[OF antiframe_prog_monoI]
  apply (rule order_trans)
  oops

lemma
  assumes vwb_lens: \<open>vwb_lens G\<close>
  assumes modifies: \<open>G:[body] = body\<close>
  shows \<open>G:[WHILE b DO body OD] = WHILE b DO body OD\<close>
  unfolding while_lfp_prog_mu_prog lfp_prog_alt_def
  apply (rule antisym)
   apply (rule inf_prog_lower)
   apply auto
   apply (subst lfp_prog_alt_def[THEN sym])
   apply (subst  lfp_prog_unfold)
    apply (fold lfp_prog_alt_def)
  apply (simp add: Mono_progI if_prog_monoI seq_prog_monoI)
   apply (subst If_prog_anti_frame_prog_distr[THEN sym])
   apply (subst seq_prog_anti_frame_prog_distr[THEN sym])
     defer 
     defer
     apply (unfold modifies)
     apply (subst algebraic_laws_prog.SKIP_prog_anti_frame_prog_modifies)
      defer
      apply simp
  unfolding lfp_prog_alt_def
     apply (subst anti_frame_is_continues)
      defer
   
  apply (rule lfp_prog_greatest)
  thm lfp_prog_greatest
   apply auto
   apply (subst anti_frame_is_continues[THEN sym])
  thm monoI
  oops

end


